from typing import Dict, List, Optional, Tuple
import warnings

from matplotlib import cm, pyplot as plt
import matplotlib.dates as mdates
import pandas as pd
import numpy as np
import seaborn as sns

from macrosynergy.management.utils.df_utils import (
    reduce_df,
    update_df,
    _map_to_business_day_frequency,
)
from macrosynergy.management.types import QuantamentalDataFrame
from macrosynergy.panel import linear_composite, make_zn_scores


class ScoreVisualisers:
    """
    Class to visualize the z-scores and linear composite of specified categories and
    cross-sections.

    Parameters
    ----------
    df : ~pandas.DataFrame
        A standardized JPMaQS with the following columns: 'cid', 'xcat', 'real_date',
        and at least one metric from - 'value', 'grading', 'eop_lag', or 'mop_lag'.
    cids : List[str], optional
        A list of cross-section identifiers to select from the DataFrame. If None, all
        cross-sections in the frame are selected. Default is None.
    xcats : List[str], optional
        A list of category tickers to select from the DataFrame. If None, all categories
        are selected. Default is None.
    xcat_labels : Dict[str, str], optional
        A dictionary mapping category tickers (keys) to their labels (values). Default is
        None.
    xcat_comp : str
        The name of the composite category. Default is 'Composite'.
    weights : List[float]
        A list of weights for the linear composite. Default is equal weights. The length
        of the list must be equal to the number of categories in xcats. If weights do not
        add up to 1, they are normalized.
    normalize_weights : bool
        If True (default), normalize weights if they do not add to one.
    signs : List[float]
        A list of signs in order to use both negative and positive values of categories
        for the linear composite. This must have the same length as weights and xcats, and
        correspondes to the order of categories in xcats. Default is all positive.
    blacklist : Dict[str, str]
        A dictionary of cross-sections (keys) and date ranges (values) that should be
        excluded. If one cross-section has several blacklist periods append numbers to the
        cross-section identifier.
    complete_xcats : bool
        If True, all xcats must have data for the composite to be calculated. Default is
        False, which means that the composite is calculate if at least one category has
        data.
    no_zn_scores : bool
        Per default, all categories are scored before they are averaged into the
        composite. If True, the class does not calculate scores and takes the average of the
        original categiries. This is useful if those are already score or of similar scale.
    rescore_composite : bool
        If True, the composite is re-scored to a normal unit scale. Default is False.
    sequential : bool
        if True (default) score parameters (neutral level and mean absolute deviation)
        are estimated sequentially with concurrently available information only.
    min_obs : int
        the minimum number of observations required to calculate zn_scores. Default is
        261. The parameter is only applicable if the "sequential" parameter is set to True.
        Otherwise the neutral level and the mean absolute deviation are both computed in-
        sample and will use the full sample.
    iis : bool
        if True (default) zn-scores are also calculated for the initial sample period
        defined by min-obs on an in-sample basis to avoid losing history. This is irrelevant
        if sequential is set to False.
    neutral : str
        The method to calculate the neutral score. Default is 'zero'. Alternatives are
        'mean', 'median' or a number.
    pan_weight : float
        The weight of panel (versus individual cross section) for calculating the
        z-score parameters, i.e. the neutral level and the mean absolute deviation. Default
        is 1, i.e. panel data are the basis for the parameters. Lowest possible value is 0,
        i.e. parameters are all specific to cross section.
    thresh : float
        The threshold value beyond which scores are winsorized, i.e. contained at that
        threshold. The threshold is the maximum absolute score value that the function is
        allowed to produce. The minimum threshold is 1 mean absolute deviation.
    est_freq : str
        the frequency at which mean absolute deviations or means are re-estimated. The
        options are daily, weekly, monthly & quarterly: "D", "W", "M", "Q". Default is
        monthly. Re-estimation is performed at period end.
    postfix : str
        The string appended to category name for output; default is "_ZN".
    """

    def __init__(
        self,
        df: pd.DataFrame,
        xcats: List[str] = None,
        cids: List[str] = None,
        xcat_labels: Dict[str, str] = None,
        xcat_comp: str = "Composite",
        weights: List[float] = None,
        blacklist: Dict[str, str] = None,
        sequential: bool = True,
        iis: bool = True,
        neutral: str = "mean",
        pan_weight: float = 1,
        thresh: float = 3,
        min_obs: int = 261,
        est_freq: str = "m",
        postfix: str = "_ZN",
        normalize_weights: bool = True,
        signs: Optional[List[float]] = None,
        complete_xcats: bool = False,
        no_zn_scores: bool = False,
        rescore_composite: bool = False,
    ):
        self._validate_params(cids, xcats, xcat_comp)

        self.cids = cids if cids else list(df["cid"].unique())
        self.weights = weights
        self.postfix = postfix
        if no_zn_scores:
            self.postfix = ""
        self.xcat_comp = xcat_comp + self.postfix
        df = QuantamentalDataFrame(df)
        self.df = self._create_df(
            df,
            xcats,
            blacklist,
            sequential,
            iis,
            neutral,
            pan_weight,
            thresh,
            min_obs,
            est_freq,
            postfix,
            no_zn_scores,
        )
        self.old_xcats = [xcat_comp] + xcats
        self.xcats = self._apply_postfix(xcats)
        for xcat in self.xcats:
            if xcat not in self.df["xcat"].unique():
                self.xcats.remove(xcat)
                warnings.warn(f"{xcat} not in the DataFrame")

        composite_df = linear_composite(
            self.df,
            xcats=self.xcats,
            cids=self.cids,
            weights=self.weights,
            normalize_weights=normalize_weights,
            signs=signs,
            blacklist=blacklist,
            complete_xcats=complete_xcats,
            new_xcat=self.xcat_comp,
        )

        if rescore_composite:
            composite_df = make_zn_scores(
                composite_df,
                xcat=self.xcat_comp,
                sequential=sequential,
                cids=self.cids,
                blacklist=blacklist,
                iis=iis,
                neutral=neutral,
                pan_weight=pan_weight,
                thresh=thresh,
                min_obs=min_obs,
                est_freq=est_freq,
                postfix="",
            )

        self.df = update_df(self.df, composite_df)
        self.xcats = [self.xcat_comp] + self.xcats
        self.xcat_labels = xcat_labels

    def _validate_params(self, cids, xcats, xcat_comp):
        if cids and (
            not isinstance(cids, list) or not all(isinstance(cid, str) for cid in cids)
        ):
            raise TypeError("cids must be a list of strings")

        if xcats and (
            not isinstance(xcats, list)
            or not all(isinstance(xcat, str) for xcat in xcats)
        ):
            raise TypeError("xcats must be a list of strings")

        if not isinstance(xcat_comp, str):
            raise TypeError("xcat_comp must be a string")

    def _create_df(
        self,
        df,
        xcats,
        blacklist,
        sequential,
        iis,
        neutral,
        pan_weight,
        thresh,
        min_obs,
        est_freq,
        postfix,
        no_zn_scores,
    ):
        """
        Helper function to create the DataFrame with z-scores.
        """
        if no_zn_scores:
            return reduce_df(df, xcats=xcats, cids=self.cids, blacklist=blacklist)

        result_df = None
        for xcat in xcats:
            dfzm = make_zn_scores(
                df,
                xcat=xcat,
                sequential=sequential,
                cids=self.cids,
                blacklist=blacklist,
                iis=iis,
                neutral=neutral,
                pan_weight=pan_weight,
                thresh=thresh,
                min_obs=min_obs,
                est_freq=est_freq,
                postfix=postfix,
            )
            result_df = update_df(result_df, dfzm) if result_df is not None else dfzm
        return result_df

    def _plot_heatmap(
        self,
        df: pd.DataFrame,
        title: str,
        title_fontsize: int = 20,
        annot: bool = True,
        xticks=None,
        yticks_rotation: Optional[int] = None,
        figsize=(20, 10),
        round_decimals: int = 2,
        cmap: str = None,
        cmap_range: Tuple[float, float] = None,
        horizontal_divider: bool = False,
        vertical_divider: bool = False,
        divider_position: int = None,
    ):
        """
        Helper function to plot the heatmap.
        """
        fig, ax = plt.subplots(figsize=figsize)

        cmap = cmap or "coolwarm_r"
        vmax = np.nanmax(np.abs(df.values))
        vmin = -vmax
        cmap_range = cmap_range or (vmin, vmax)

        sns.heatmap(
            df,
            cmap=cmap,
            annot=annot,
            xticklabels="auto",
            yticklabels="auto",
            fmt=f".{round_decimals}f",
            ax=ax,
            vmin=cmap_range[0],
            vmax=cmap_range[1],
        )

        ax.set_title(title, fontsize=title_fontsize)

        if horizontal_divider:
            ax.hlines([divider_position], *ax.get_xlim(), linewidth=2, color="black")
        if vertical_divider:
            ax.vlines([divider_position], *ax.get_ylim(), linewidth=2, color="black")

        plt.xticks(**(xticks or {"rotation": 45, "ha": "right"}))
        if yticks_rotation is not None:
            plt.yticks(rotation=yticks_rotation)
        plt.tight_layout()
        plt.show()

    def _apply_postfix(self, items: List[str]) -> List[str]:
        return [
            item + self.postfix if not item.endswith(self.postfix) else item
            for item in items
        ]

    def _strip_postfix(self, items: List[str]) -> List[str]:
        return [
            item[: -len(self.postfix)] if item.endswith(self.postfix) else item
            for item in items
        ]

    def view_snapshot(
        self,
        cids: List[str] = None,
        xcats: List[str] = None,
        transpose: bool = False,
        date: str = None,
        annot: bool = True,
        title: str = None,
        title_fontsize: int = 20,
        figsize: tuple = (20, 10),
        xcat_labels: Dict[str, str] = None,
        xticks: dict = None,
        yticks_rotation: Optional[int] = None,
        round_decimals: int = 2,
        cmap: str = None,
        cmap_range: Tuple[float, float] = None,
        sort_by_composite: bool = False,
        composite_to_end: bool = False,
    ):
        """
        View heatmap of the scores at the specified or latest available date.

        Parameters
        ----------
        cids : List[str]
            A list of cross-section identifiers to select from the DataFrame. If None,
            all cross-sections in the frame are selected.
        xcats : List[str]
            A list of category tickers to select from the DataFrame. If None, all
            categories are selected.
        transpose : bool
            If True, transpose the snapshot so cross-section identifiers are on the
            x-axis and category tickers are on the y-axis.
        date : str
            ISO-8601 formatted date. The date of the snapshot. If None, the latest date
            in the DataFrame is selected. Unless the date is today, then the latest date is
            set to the previous business day.
        annot : bool
            If True, annotate the heatmap.
        title : str
            The title of the heatmap.
        title_fontsize : int
            The fontsize of the title.
        figsize : tuple
            The size of the figure.
        xcat_labels : dict
            A dictionary mapping category tickers to their labels.
        xticks : dict
            A dictionary of arguments to label the x axis.
        yticks_rotation : int
            The rotation of the y-axis labels.
        round_decimals : int
            The number of decimals to round the scores to.
        cmap : str
            The colormap of the heatmap.
        cmap_range : tuple
            The range of the colormap.
        """

        cids = cids or self.cids
        xcats = xcats or self.xcats
        xcats = self._apply_postfix(xcats)
        xcat_labels = xcat_labels or self.xcat_labels

        if date:
            date = pd.to_datetime(date)
        else:
            if (
                self.df["real_date"].max().normalize()
                == pd.Timestamp.today().normalize()
            ):
                date = pd.Timestamp.today() - pd.tseries.offsets.BDay(1)
            else:
                date = self.df["real_date"].max()

        date = date.strftime("%Y-%m-%d")

        df = self.df[
            (self.df["xcat"].isin(xcats))
            & (self.df["cid"].isin(cids))
            & (self.df["real_date"] == date)
        ]
        dfw = df.pivot(index="cid", columns="xcat", values="value")
        dfw = dfw.reindex(cids)
        dfw.columns.name = None
        dfw.index.name = None

        composite_zscore = self.xcat_comp

        if sort_by_composite:
            if composite_zscore not in xcats:
                raise ValueError(
                    "Composite category must be in xcats to sort by composite"
                )
            dfw = dfw.sort_values(by=composite_zscore, ascending=False)
        if composite_zscore in xcats:
            if composite_to_end:
                ordering = [xcat for xcat in xcats if xcat != composite_zscore] + [composite_zscore]
            else:
                ordering = [composite_zscore] + [xcat for xcat in xcats if xcat != composite_zscore]
            dfw = dfw[ordering]
        else:
            dfw = dfw[xcats]

        if xcat_labels:
            dfw.columns = [
                xcat_labels.get(
                    self._strip_postfix([xcat])[0],
                    xcat_labels.get(self._apply_postfix([xcat])[0], xcat),
                )
                for xcat in dfw.columns
            ]

        # Drop columns and rows with all NaNs
        dfw = dfw.dropna(axis=1, how="all")
        dfw = dfw.dropna(axis=0, how="all")

        if transpose:
            dfw = dfw.transpose()

        title = title or f"Snapshot for {date}"

        horizontal_divider = transpose and composite_zscore in xcats
        vertical_divider = not transpose and composite_zscore in xcats
        divider_position = None
        if composite_zscore in xcats:
            if composite_to_end:
                divider_position = len(xcats) - 1
            else:
                divider_position = 1

        self._plot_heatmap(
            dfw,
            title=title,
            annot=annot,
            xticks=xticks,
            yticks_rotation=yticks_rotation,
            figsize=figsize,
            title_fontsize=title_fontsize,
            round_decimals=round_decimals,
            cmap=cmap,
            cmap_range=cmap_range,
            horizontal_divider=horizontal_divider,
            vertical_divider=vertical_divider,
            divider_position=divider_position,
        )

    def view_score_evolution(
        self,
        xcat: str,
        freq: str = "A",
        cids: List[str] = None,
        include_latest_period: bool = True,
        include_latest_day: bool = True,
        start: str = None,
        transpose: bool = False,
        annot: bool = True,
        title: str = None,
        title_fontsize: int = 20,
        xticks: dict = None,
        figsize: tuple = (20, 10),
        cmap: str = None,
        cmap_range: Tuple[float, float] = None,
        round_decimals: int = 2,
    ):
        """
        View the evolution of the scores for the specified category and cross-sections.

        Parameters
        ----------
        xcat : str
            The category to view the evolution of.
        freq : str
            The frequency of the evolution. Must be 'Q', 'A', or 'BA'.
        cids : List[str]
            A list of cross-section identifiers to select from the DataFrame. If None,
            all in the dataframe are selected.
        include_latest_period : bool
            If True, include the latest period in the evolution.
        include_latest_day : bool
            If True, include the latest day in the evolution. If the latest date is
            today, then the latest date is set to the previous business day.
        date : str
            ISO-8601 formatted date. The date of the snapshot. If None, the latest date
            in the DataFrame is selected.
        start : str
            ISO-8601 formatted date. Select data from this date onwards. If None, all
            dates are selected.
        transpose : bool
            If True, transpose the evolution so cross-section identifiers are on the
            x-axis and dates are on the y-axis.
        annot : bool
            If True, annotate the heatmap.
        title : str
            The title of the heatmap.
        title_fontsize : int
            The fontsize of the title.
        xticks : dict
            A dictionary of arguments to label the x axis.
        figsize : tuple
            The size of the figure.
        round_decimals : int
            The number of decimals to round the scores to.
        cmap : str
            The colormap of the heatmap.
        cmap_range : tuple
            The range of the colormap.
        """

        cids = cids or self.cids
        xcat = xcat if xcat.endswith(self.postfix) else xcat + self.postfix

        freq = "2AS" if freq == "BA" else _map_to_business_day_frequency(freq)

        if not (freq in ["2AS", "BA", "A"] or freq.startswith("BQ")):
            raise ValueError("freq must be 'Q', 'A', or 'BA'")

        df = self.df[self.df["xcat"] == xcat]
        df = df[df["cid"].isin(cids)]
        df = df if start is None else df[df["real_date"] >= start]

        dfw = df.pivot(index="real_date", columns="cid", values="value")
        dfw.columns.name = None
        dfw.index.name = None
        dfw_resampled = dfw.resample(freq, origin="start").mean()

        if not include_latest_period:
            dfw_resampled = dfw_resampled.iloc[:-1]

        if include_latest_day:
            if (
                df["real_date"].max().normalize()
                == pd.Timestamp.today().normalize()
            ):
                dfw_resampled.loc[
                    df["real_date"].max() - pd.tseries.offsets.BDay(1)
                ] = dfw.loc[
                    df["real_date"].max() - pd.tseries.offsets.BDay(1)
                ]
                print(
                    "Latest day: ",
                    df["real_date"].max() - pd.tseries.offsets.BDay(1),
                )
            else:
                dfw_resampled.loc[df["real_date"].max()] = dfw.loc[
                    df["real_date"].max()
                ]
                print("Latest day: ", df["real_date"].max())
            if freq in ["Q", "BQ", "BQE"]:
                dfw_resampled.index = list(
                    dfw_resampled.index.to_period("Q").strftime("%YQ%q")[:-1]
                ) + ["Latest"]
            else:
                dfw_resampled.index = list(dfw_resampled.index.strftime("%Y")[:-1]) + [
                    "Latest"
                ]
        else:
            if freq in ["Q", "BQ", "BQE"]:
                dfw_resampled.index = list(
                    dfw_resampled.index.to_period("Q").strftime("%YQ%q")
                )
            else:
                dfw_resampled.index = list(dfw_resampled.index.strftime("%Y"))

        dfw_resampled = dfw_resampled.transpose()
        dfw_resampled = dfw_resampled.reindex(cids)

        # Drop columns and rows with all NaNs
        dfw_resampled = dfw_resampled.dropna(axis=1, how="all")
        dfw_resampled = dfw_resampled.dropna(axis=0, how="all")

        if transpose:
            dfw_resampled = dfw_resampled.transpose()

        title = title or f"Evolution for {xcat}"

        self._plot_heatmap(
            dfw_resampled,
            title=title,
            annot=annot,
            xticks=xticks,
            figsize=figsize,
            title_fontsize=title_fontsize,
            round_decimals=round_decimals,
            cmap=cmap,
            cmap_range=cmap_range,
        )

    def view_cid_evolution(
        self,
        cid: str,
        xcats: List[str],
        freq: str,
        include_latest_period: bool = True,
        include_latest_day: bool = True,
        start: str = None,
        transpose: bool = False,
        annot: bool = True,
        title: str = None,
        title_fontsize: int = 20,
        figsize: tuple = (20, 10),
        xticks: dict = None,
        xcat_labels: Dict[str, str] = None,
        cmap: str = None,
        cmap_range: Tuple[float, float] = None,
        round_decimals: int = 2,
        composite_to_end: bool = False,
    ):
        """
        View the evolution of the scores for the specified cross-section and categories.

        Parameters
        ----------
        cid : str
            The cross-section to view the evolution of.
        xcats : List[str]
            A list of category tickers to select from the DataFrame. If None, all are
            selected.
        freq : str
            The frequency of the evolution. Must be 'Q', 'A', or 'BA'.
        include_latest_period : bool
            If True, include the latest period in the evolution.
        include_latest_day : bool
            If True, include the latest day in the evolution. If the latest date is
            today, then the latest date is set to the previous business day.
        start : str
            ISO-8601 formatted date. Select data from this date onwards. If None, all
            dates are selected.
        transpose : bool
            If True, transpose the evolution so xcats are on the x-axis and dates are on
            the y-axis.
        annot : bool
            If True, annotate the heatmap.
        title : str
            The title of the heatmap.
        title_fontsize : int
            The fontsize of the title.
        xticks : dict
            A dictionary of arguments to label the x axis.
        figsize : tuple
            The size of the figure.
        xcat_labels : dict
            A dictionary mapping xcats to their labels.
        round_decimals : int
            The number of decimals to round the scores to.
        cmap : str
            The colormap of the heatmap.
        cmap_range : tuple
            The range of the colormap.
        """

        if not isinstance(cid, str):
            raise TypeError("cid must be a string")

        freq = "2AS" if freq == "BA" else _map_to_business_day_frequency(freq)

        if not (freq in ["2AS", "BA", "A"] or freq.startswith("BQ")):
            raise ValueError("freq must be 'Q', 'A', or 'BA'")

        xcat_labels = xcat_labels or self.xcat_labels
        xcats = self._apply_postfix(xcats)

        df = self.df[self.df["cid"] == cid]
        df = df if start is None else df[df["real_date"] >= start]
        df = df[df["xcat"].isin(xcats)]
        
        # If there is an xcat that does not exist in the DataFrame, remove it and warn
        for xcat in xcats:
            if xcat not in df["xcat"].unique():
                xcats.remove(xcat)
                warnings.warn(f"{xcat} not in the DataFrame")

        dfw = df.pivot(index="real_date", columns="xcat", values="value")
        dfw.columns.name = None
        dfw.index.name = None
        dfw_resampled = dfw.resample(freq).mean()

        if not include_latest_period:
            dfw_resampled = dfw_resampled.iloc[:-1]

        if include_latest_day:
            if (
                df["real_date"].max().normalize()
                == pd.Timestamp.today().normalize()
            ):
                dfw_resampled.loc[
                    df["real_date"].max() - pd.tseries.offsets.BDay(1)
                ] = dfw.loc[
                    df["real_date"].max() - pd.tseries.offsets.BDay(1)
                ]
                print(
                    "Latest day: ",
                    df["real_date"].max() - pd.tseries.offsets.BDay(1),
                )
            else:
                dfw_resampled.loc[df["real_date"].max()] = dfw.loc[
                    df["real_date"].max()
                ]
                print("Latest day: ", df["real_date"].max())
            if freq in ["Q", "BQ", "BQE"]:
                dfw_resampled.index = list(
                    dfw_resampled.index.to_period("Q").strftime("%YQ%q")[:-1]
                ) + ["Latest"]
            else:
                dfw_resampled.index = list(dfw_resampled.index.strftime("%Y")[:-1]) + [
                    "Latest"
                ]
        else:
            if freq in ["Q", "BQ", "BQE"]:
                dfw_resampled.index = list(
                    dfw_resampled.index.to_period("Q").strftime("%YQ%q")
                )
            else:
                dfw_resampled.index = list(dfw_resampled.index.strftime("%Y"))

        composite_zscore = self.xcat_comp
        if composite_zscore in xcats:
            dfw_resampled = dfw_resampled[
                [composite_zscore]
                + [xcat for xcat in xcats if xcat != composite_zscore]
            ]
        else:
            dfw_resampled = dfw_resampled[xcats]

        if xcat_labels:
            dfw_resampled.columns = [
                xcat_labels.get(
                    self._strip_postfix([xcat])[0],
                    xcat_labels.get(self._apply_postfix([xcat])[0], xcat),
                )
                for xcat in dfw_resampled.columns
            ]

        dfw_resampled = dfw_resampled.transpose()

        # Drop columns and rows with all NaNs
        dfw_resampled = dfw_resampled.dropna(axis=1, how="all")
        dfw_resampled = dfw_resampled.dropna(axis=0, how="all")

        if transpose:
            dfw_resampled = dfw_resampled.transpose()

        title = title or f"Evolution for {cid}"

        horizontal_divider = not transpose and self.xcat_comp in xcats
        vertical_divider = transpose and self.xcat_comp in xcats
        divider_position = None
        if composite_zscore in xcats:
            if composite_to_end:
                divider_position = len(xcats) - 1
            else:
                divider_position = 1

        self._plot_heatmap(
            dfw_resampled,
            title=title,
            annot=annot,
            xticks=xticks,
            figsize=figsize,
            title_fontsize=title_fontsize,
            round_decimals=round_decimals,
            cmap=cmap,
            cmap_range=cmap_range,
            horizontal_divider=horizontal_divider,
            vertical_divider=vertical_divider,
            divider_position=divider_position,
        )


if __name__ == "__main__":
    cids_dmea = ["FRF", "DEM", "ITL", "ESP", "EUR"]
    cids_dmxe = ["CHF", "GBP", "JPY", "SEK", "USD"]
    cids_dm = cids_dmea + cids_dmxe
    cids_g10 = ["AUD", "DEM", "FRF", "ESP", "ITL", "JPY", "NZD", "GBP", "USD"]
    cids_latm = ["BRL", "CLP", "COP", "MXN", "PEN"]  # Latam sovereigns
    cids_emea = [
        "CZK",
        "HUF",
        "ILS",
        "PLN",
        "RON",
        "ZAR",
        "TRY",
    ]  # EMEA sovereigns
    cids_emas = [
        "CNY",
        "IDR",
        "KRW",
        "MYR",
        "PHP",
        "THB",
    ]  # EM Asia sovereigns
    cids_ea = ["DEM", "FRF", "ESP", "ITL"]  # major Euro currencies before EUR
    cids_em = cids_emea + cids_latm + cids_emas
    cids = cids_dm + cids_em
    main = [
        "GGIEDGDP_NSA",
        # Currency reserve expansion as % of GDP
        "NIIPGDP_NSA",
        # Monetary base expansion as % of GDP
        "CABGDPRATIO_NSA_12MMA",
        # Intervention-driven liquidity expansion as % of GDP, diff over 3 months
        "GGOBGDPRATIO_NSA",
        # Intervention-driven liquidity expansion as % of GDP, diff over 6 months
    ]

    rets = []

    # rets = [
    #     "DU05YXR_NSA",
    #     "DU05YXR_VT10",
    #     "EQXR_NSA",
    #     "EQXR_VT10",
    #     "FXTARGETED_NSA",
    #     "FXUNTRADABLE_NSA",
    #     "FXXR_VT10",
    # ]

    xcats = main + rets

    tickers = [cid + "_" + xcat for cid in cids for xcat in xcats]

    start_date = "1990-01-01"

    import os
    from macrosynergy.download import JPMaQSDownload

    # Retrieve credentials
    client_id: str = os.getenv("DQ_CLIENT_ID")
    client_secret: str = os.getenv("DQ_CLIENT_SECRET")

    with JPMaQSDownload(client_id=client_id, client_secret=client_secret) as dq:
        df = dq.download(
            tickers=tickers,
            start_date=start_date,
            suppress_warning=True,
            metrics=["all"],
            show_progress=True,
        )

    # Remove data for 24th Feb 2025 for "GGIEDGDP_NSA"

    df = df[
        ~(
            (df["cid"] == "USD")
            & (df["xcat"] == "GGIEDGDP_NSA")
            & (df["real_date"] == "2025-02-24")
        )
    ]

    blacklist = {
        "USD": [pd.Timestamp("2020-06-06"), pd.Timestamp("2030-07-23")]
    }

    sv = ScoreVisualisers(
        df,
        cids=cids,
        xcats=xcats,
        thresh=3,
        no_zn_scores=True,
        complete_xcats=False,
        rescore_composite=True,
        blacklist=blacklist,
    )

    sv.view_snapshot(
        cids=["USD", "EUR", "JPY", "GBP", "CHF"],
        xcats=xcats + ["Composite"],
        figsize=(14, 12),
        sort_by_composite=True,
        composite_to_end=True,
        transpose=False,
        yticks_rotation=45,
    )
    sv.view_cid_evolution(cid="USD", xcats=xcats + ["Composite"] , freq="A", transpose=False)
    sv.view_score_evolution(
        xcat="GGIEDGDP_NSA",
        cids=cids,
        freq="BA",
        transpose=False,
        start="2010-01-01",
        title="AHKSJDA",
        include_latest_day=True,
    )



"""
Module containing code for the Plotter class, and all common functionality for the
plotter classes. The Plotter class is a base class for all plotter classes, and provides
a shared interface for dataframe filtering operations, as well as `argvalidation` and
`argcopy` decorators for all methods of the plotter classes.
"""

import logging
from types import ModuleType
from typing import Any, Dict, List, Optional

import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd

from macrosynergy.management.decorators import argcopy, argvalidation
from macrosynergy.management.validation import validate_and_reduce_qdf
from macrosynergy.management.types import QuantamentalDataFrame

logger = logging.getLogger(__name__)


class PlotterMetaClass(type):
    """
    Metaclass for the Plotter class. The purpose of this metaclass is to wrap all
    methods of the Plotter class with the `argvalidation` and `argcopy` decorators, so
    that all methods of the Plotter class are automatically validated and copied. Meant
    to be used as a metaclass, i.e. use as follows:

    .. code-block:: python

        class MyCustomClass(metaclass=PlotterMetaClass):
            def __init__(self, ...):
                ...
    """

    def __init__(cls, name, bases, dct: Dict[str, Any]):
        super().__init__(name, bases, dct)
        for attr_name, attr_value in dct.items():
            if callable(attr_value):
                setattr(cls, attr_name, argcopy(argvalidation(attr_value)))


class Plotter(metaclass=PlotterMetaClass):
    """
    Base class for a DataFrame Plotter. The inherited meta class automatically wraps all
    methods of the Plotter class and any subclasses with the `argvalidation` and
    `argcopy` decorators, so that all methods of the class are automatically validated
    and copied. This class does not implement any plotting functionality, but provides a
    shared interface for the plotter classes, and some common functionality - currently
    just the filtering of the DataFrame.

    Parameters
    ----------
    df : ~pandas.DataFrame
        A DataFrame with the following columns: 'cid', 'xcat', 'real_date', and at least
        one metric from - 'value', 'grading', 'eop_lag', or 'mop_lag'.
    cids : List[str]
        A list of cids to select from the DataFrame. If None, all cids are selected.
    xcats : List[str]
        A list of xcats to select from the DataFrame. If None, all xcats are selected.
    metrics : List[str]
        A list of metrics to select from the DataFrame. If None, all metrics are
        selected.
    intersect : bool
        if True only retains cids that are available for all xcats. Default is False.
    tickers : List[str]
        A list of tickers that will be selected from the DataFrame if they exist,
        regardless of start, end, blacklist, and intersect arguments.
    blacklist : dict
        cross-sections with date ranges that should be excluded from the data frame. If
        one cross-section has several blacklist periods append numbers to the cross-section
        code.
    start : str
        ISO-8601 formatted date string. Select data from this date onwards. If None, all
        dates are selected.
    end : str
        ISO-8601 formatted date string. Select data up to and including this date. If
        None, all dates are selected.
    backend : str
        The plotting backend to use. Currently only 'matplotlib' is supported.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        cids: Optional[List[str]] = None,
        xcats: Optional[List[str]] = None,
        metrics: Optional[List[str]] = None,
        intersect: Optional[bool] = False,
        tickers: Optional[List[str]] = None,
        blacklist: Optional[Dict[str, List[str]]] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        backend: Optional[str] = "matplotlib",
    ):
        validated_args = validate_and_reduce_qdf(
            df=df,
            cids=cids,
            xcats=xcats,
            metrics=metrics,
            intersect=intersect,
            tickers=tickers,
            blacklist=blacklist,
            start=start,
            end=end,
        )

        self.df: pd.DataFrame = QuantamentalDataFrame(validated_args.df)
        self.cids: List[str] = validated_args.cids
        self.xcats: List[str] = validated_args.xcats
        self.metrics: List[str] = validated_args.metrics
        self.intersect: bool = intersect
        self.tickers: List[str] = tickers
        self.blacklist: Dict[str, List[str]] = blacklist
        self.start: str = validated_args.start
        self.end: str = validated_args.end

        self.backend: ModuleType
        accepted_backends: List[str] = ["matplotlib", "plt", "mpl"]
        if not backend.strip().lower() in accepted_backends:
            raise NotImplementedError(f"Backend `{backend}` is not supported.")

        self.backend = plt
        sns.set_theme(style="darkgrid", palette="colorblind")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass



"""
Module for calculating notional positions based on contract signals, assets-under-
management, and other relevant parameters.
"""

import numpy as np
import pandas as pd
from numbers import Number
from typing import List, Union, Tuple, Optional, Set

from macrosynergy.management.utils import (
    standardise_dataframe,
    reduce_df,
    is_valid_iso_date,
    apply_slip as apply_slip_util,
    ticker_df_to_qdf,
)

from macrosynergy.management.types import NoneType, QuantamentalDataFrame
from macrosynergy.pnl.historic_portfolio_volatility import historic_portfolio_vol


def _apply_slip(
    df: pd.DataFrame,
    slip: int,
    fids: List[str],
) -> pd.DataFrame:
    """
    Applies a slip using the function `apply_slip()` to a dataframe with contract
    signals and returns.

    Parameters
    ----------
    df : pd.DataFrame
        Quantamental dataframe with contract signals and returns.
    slip : int
        the number of days to wait before applying the signal.
    fids : List[str]
        list of contract identifiers to apply the slip to.
    metrics : List[str]
        list of metrics to apply the slip to.
    """

    assert isinstance(df, QuantamentalDataFrame)
    assert isinstance(slip, int)
    assert (
        isinstance(fids, list)
        and len(fids) > 0
        and all([isinstance(x, str) for x in fids])
    )

    if slip == 0:
        return df
    else:
        cdf = QuantamentalDataFrame(df)
        filter_tickers = [tk for tk in cdf.list_tickers() if tk.startswith(tuple(fids))]
        cdf = cdf.reduce_df_by_ticker(tickers=filter_tickers)
        return apply_slip_util(
            df=cdf,
            tickers=filter_tickers,
            slip=slip,
            raise_error=False,
            metrics=["value"],
            extend_dates=True,
        )


def _check_df_for_contract_signals(
    df_wide: pd.DataFrame,
    sname: str,
    fids: List[str],
) -> None:
    """
    Checks if the dataframe contains contract signals for the specified strategy and the
    specified contract identifiers.

    Parameters
    ----------
    df : pd.DataFrame
        Wide dataframe with contract signals and returns.
    sname : str
        the name of the strategy.
    fids : List[str]
        list of contract identifiers to apply the slip to.
    """

    assert isinstance(sname, str)
    assert (
        isinstance(fids, list)
        and len(fids) > 0
        and all([isinstance(x, str) for x in fids])
    )

    sig_ident: str = f"_CSIG_{sname}"
    _check_conts: Set = set([f"{contx}{sig_ident}" for contx in fids])
    _found_conts: Set = set(df_wide.columns)
    if not _check_conts.issubset(_found_conts):
        raise ValueError(
            f"Contract signals for all contracts not in dataframe. \n"
            f"Missing: {_check_conts - _found_conts}"
        )

    return


def _vol_target_positions(
    df_wide: pd.DataFrame,
    sname: str,
    fids: List[str],
    aum: Number,
    vol_target: Number,
    rebal_freq: str,
    est_freqs: Union[str, List[str]],
    est_weights: Union[Number, List[Number]],
    lback_periods: Union[int, List[int]],
    half_life: Union[int, List[int]],
    nan_tolerance: float,
    remove_zeros: bool,
    lback_meth: str,
    rstring: str,
    pname: str,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Uses historic portfolio volatility to calculate notional positions based on contract
    signals, volatility targeting and other relevant parameters.
    """

    _check_df_for_contract_signals(df_wide=df_wide, sname=sname, fids=fids)

    sig_ident: str = f"_CSIG_{sname}"

    # TODO what is the units of histpvol?
    histpvol: QuantamentalDataFrame
    vcv_df: pd.DataFrame
    histpvol, vcv_df = historic_portfolio_vol(
        df=ticker_df_to_qdf(df_wide),
        sname=sname,
        fids=fids,
        rstring=rstring,
        lback_meth=lback_meth,
        est_freqs=est_freqs,
        est_weights=est_weights,
        lback_periods=lback_periods,
        half_life=half_life,
        nan_tolerance=nan_tolerance,
        rebal_freq=rebal_freq,
        remove_zeros=remove_zeros,
        return_variance_covariance=True,
    )
    # histpvol: only on rebalance dates...
    histpvol["scale"] = ((vol_target / histpvol["value"]) * aum).replace(np.inf, np.nan)
    # TODO check inf => convert to NaN
    histpvol.set_index("real_date", inplace=True)

    out_df = pd.DataFrame(index=df_wide.index)

    signal_columns: List[str] = [f"{contx:s}{sig_ident:s}" for contx in fids]
    df_signals: pd.DataFrame = df_wide.loc[histpvol.index, signal_columns]

    out_df: pd.DataFrame = (
        histpvol[["scale"]].dot(np.ones(shape=(1, df_signals.shape[1]))).values
        * df_signals
    )
    # TODO how to deal with unbalanced panel

    # drop rows with all na
    # TODO add log statement of how many N/A values are dropped
    out_df = out_df.reindex(df_wide.index)
    rebal_dates = sorted(histpvol.index.tolist())

    for num, rb in enumerate(rebal_dates[:-1]):
        mask = (out_df.index >= rb) & (out_df.index < rebal_dates[num + 1])
        out_df.loc[mask, :] = out_df.loc[mask, :].ffill()
    mask = out_df.index >= rebal_dates[-1]
    out_df.loc[mask, :] = out_df.loc[mask, :].ffill()

    # get na values per column
    na_per_col = out_df.isna().sum()
    na_per_col = na_per_col[na_per_col > 0]
    out_df = out_df.rename(
        columns={
            col: col.replace(sig_ident, f"_{sname}_{pname}")
            for col in out_df.columns.tolist()
        },
    ).dropna(how="all")

    return (
        out_df,
        standardise_dataframe(histpvol[["cid", "xcat", "value"]]),
        vcv_df,
    )


def _leverage_positions(
    df_wide: pd.DataFrame,
    sname: str,
    fids: List[str],
    aum: Number = 100,
    leverage: Number = 1.0,
    pname: str = "POS",
) -> pd.DataFrame:
    """"""
    _check_df_for_contract_signals(df_wide=df_wide, sname=sname, fids=fids)

    sig_ident: str = f"_CSIG_{sname}"

    _contracts: List[str] = [f"{contx}{sig_ident}" for contx in fids]

    rowsums: pd.Series = df_wide.loc[:, _contracts].sum(axis=1)
    # if any of the rowsums are zero, set to NaN to avoid div by zero
    rowsums[rowsums == 0] = np.nan

    for ic, contx in enumerate(fids):
        pos_col: str = f"{contx}_{sname}_{pname}"
        cont_name: str = contx + sig_ident
        # NOTE: this should be
        # dfw_pos = dfw_sigs * aum * leverage / rowsums(dfw_sigs)
        df_wide[pos_col] = df_wide[cont_name] * aum * leverage / rowsums

    # filter df to only contain position columns
    df_wide = df_wide.loc[:, [f"{contx}_{sname}_{pname}" for contx in fids]]

    return df_wide


def notional_positions(
    df: QuantamentalDataFrame,
    sname: str,
    fids: List[str],
    aum: Number = 100,
    dollar_per_signal: Number = 1.0,
    slip: int = 1,
    leverage: Optional[Number] = None,
    vol_target: Optional[Number] = None,
    nan_tolerance: float = 0.25,
    remove_zeros: bool = True,
    rebal_freq: str = "m",
    lback_meth: str = "ma",
    est_freqs: Union[str, List[str]] = ["D", "W", "M"],
    est_weights: Union[Number, List[Number]] = [1, 1, 1],
    lback_periods: Union[int, List[int]] = [-1, -1, -1],
    half_life: Union[int, List[int]] = [11, 5, 6],
    rstring: str = "XR",
    start: Optional[str] = None,
    end: Optional[str] = None,
    blacklist: Optional[dict] = None,
    pname: str = "POS",
    return_pvol: bool = False,
    return_vcv: bool = False,
) -> Union[
    QuantamentalDataFrame,
    Tuple[QuantamentalDataFrame, QuantamentalDataFrame],
    Tuple[QuantamentalDataFrame, pd.DataFrame],
    Tuple[QuantamentalDataFrame, QuantamentalDataFrame, pd.DataFrame],
]:
    """
    Calculates contract positions based on contract signals, assets under management
    (AUM) and other specifications.

    Parameters
    ----------
    df : QuantamentalDataFrame
        standardized JPMaQS DataFrame with the necessary columns: 'cid', 'xcat',
        'real_date' and 'value'. This dataframe must contain the contract-specific signals
        and possibly related return series (for vol-targeting).
    sname : str
        the name of the strategy. It must correspond to contract signals in the
        dataframe, which have the format "<cid>_<ctype>_CSIG_<sname>", and which are
        typically calculated by the function contract_signals().
    fids : List[str]
        list of financial contract identifiers in the format "<cid>_<ctype>". It must
        correspond to contract signals in the dataframe.
    aum : float
        the assets under management in USD million (for consistency). This is basis for
        all position sizes. Default is 100.
    dollar_per_signal : float
        the amount of notional (e.g. USD) per contract signal value. Default is 1. The
        default scale is arbitrary and is merely a basis for tryouts.
    leverage : float
        the ratio of the sum of notional positions to AUM. This is the main basis for
        leveraged-based positioning. Since different contracts have different expected
        volatility and correlations this method does not control expected volatility.
        Default is None, i.e. the method is not applied.
    vol_target : float
        the target volatility of the portfolio in % of AUM (For clarity, `vol_target=10`
        means 10%). This is the main parameter for volatility-targeted positioning. That
        method estimates the annualized standard deviation of the signal-based portfolio for
        a 1 USD per signal portfolio based on past variances and covariances of the contract
        returns. The estimation is managed by the function 
        :func:`macrosynergy.pnl.historic_portfolio_vol`. Default is None, i.e. the
        volatility-targeting is not applied.
    rebal_freq : str
        the rebalancing frequency. Default is 'm' for monthly. Alternatives are 'w' for
        business weekly, 'd' for daily, and 'q' for quarterly. Contract signals are taken
        from the end of the holding period and applied to positions at the beginning of the
        next period, subject to slippage.
    slip : int
        the number of days to wait before applying the signal. Default is 1. This means
        that new positions are taken at the very end of the first trading day of the holding
        period and are the basis of PnL calculation from the second Trading day onward.
    lback_meth : str
        the method to use for the lookback period of the volatility-targeting method.
        Default is 'ma' for moving average. Alternative is "xma", for exponential moving
        average. Again this is passed through to the function 
        :func:`macrosynergy.pnl.historic_portfolio_vol`.
    est_freqs : str or list of str
        the frequencies to use for the estimation of the variance-covariance matrix.
        Default is to use ["D", "W", "M"]. This is passed through to the function
        :func:`macrosynergy.pnl.historic_portfolio_vol`.
    est_weights : float or list of float
        the weights to use for the estimation of the variance-covariance matrix. The
        weights are normalized and are used to calculate the weighted average of the
        estimated variances and covariances. Default is [1, 1, 1]. This is passed
        through to the function :func:`macrosynergy.pnl.historic_portfolio_vol`.
    lback_periods : int or List[int]
        the number of periods to use for the lookback period of the volatility-targeting
        method. Each element corresponds to the the same index in `est_freqs`. Passing a
        single element will apply the same value to all frequencies. Default is [-1], which
        means that the lookback period is the full available data for all specified
    half_life : List[int]
        number of periods in the half-life of the exponential moving average. Each
        element corresponds to the same index in `est_freqs`. Default is [11, 5, 6] (for
        daily, weekly and monthly frequencies respectively). This is passed through to the
        function :func:`macrosynergy.pnl.historic_portfolio_vol`.
    rstring : str
        a general string of the return category. This identifies the contract returns
        that are required for the volatility-targeting method, based on the category
        identifier format <cid>_<ctype><rstring>_<rstring> in accordance with JPMaQS
        conventions. Default is 'XR'.
    start : str
        the start date of the data. Default is None, which means that the start date is
        taken from the dataframe.
    end : str
        the end date of the data. Default is None, which means that the end date is
        taken from the dataframe.
    blacklist : dict
        a dictionary of contract identifiers to exclude from the calculation. Default is
        None, which means that no contracts are excluded.
    pname : str
        the name of the position. Default is 'POS'.
    return_pvol : bool
        whether to return the historic portfolio volatility. Default is False.
    return_vcv : bool
        whether to return the variance-covariance matrix. Default is False.

    Returns
    -------
    pd.DataFrame
        a standard Quantamental DataFrame with the positions for all traded contracts
        and the specified strategy in USD million. The contract signals have the following
        format "<cid>_<ctype>_<sname>_<pname>".
    """
    if not isinstance(df, QuantamentalDataFrame):
        raise ValueError("`df` must be a QuantamentalDataFrame.")

    for varx, namex, typex in [
        (sname, "sname", str),
        (fids, "fids", list),
        (aum, "aum", Number),
        (dollar_per_signal, "dollar_per_signal", Number),
        (leverage, "leverage", (Number, NoneType)),
        (vol_target, "vol_target", (Number, NoneType)),
        (rebal_freq, "rebal_freq", str),
        (slip, "slip", int),
        (lback_meth, "lback_meth", str),
        (est_freqs, "est_freqs", (str, list)),
        (est_weights, "est_weights", (Number, list)),
        (lback_periods, "lback_periods", (int, list)),
        (half_life, "half_life", (int, list)),
        (rstring, "rstring", str),
        (start, "start", (str, NoneType)),
        (end, "end", (str, NoneType)),
        (blacklist, "blacklist", (dict, NoneType)),
        (pname, "pname", str),
    ]:
        if not isinstance(varx, typex):
            raise ValueError(f"`{namex}` must be {typex}.")

        if isinstance(varx, (str, list, dict)) and len(varx) == 0:
            raise ValueError(f"`{namex}` must not be an empty {str(typex)}.")

    ## Convert df to QDF
    df: QuantamentalDataFrame = QuantamentalDataFrame(df)
    _initialized_as_categorical: bool = df.InitializedAsCategorical

    ## Volatility targeting and leverage cannot be applied at the same time
    if bool(leverage) and bool(vol_target):
        raise ValueError(
            "Either `leverage` or `vol_target` must be specified, but not both."
        )

    ## Check the dates
    if start is None:
        start: str = pd.Timestamp(df["real_date"].min()).strftime("%Y-%m-%d")
    if end is None:
        end: str = pd.Timestamp(df["real_date"].max()).strftime("%Y-%m-%d")

    for dx, nx in [(start, "start"), (end, "end")]:
        if not is_valid_iso_date(dx):
            raise ValueError(f"`{nx}` must be a valid ISO-8601 date string")

    ## Reduce the dataframe
    df: QuantamentalDataFrame = reduce_df(
        df=df, start=start, end=end, blacklist=blacklist
    )

    ## Check the contract identifiers and contract signals

    # df["ticker"] = df["cid"] + "_" + df["xcat"]
    df = QuantamentalDataFrame(df).add_ticker_column()

    # There must be atleast one contract signal with the strategy name
    if not any(df["ticker"].str.endswith(f"_CSIG_{sname}")):
        raise ValueError(f"No contract signals for strategy `{sname}` in dataframe.")

    # Check that all contract identifiers have at least one signal
    u_tickers: List[str] = list(df["ticker"].unique())
    for contx in fids:
        if not any(
            [tx.startswith(contx) and tx.endswith(f"_CSIG_{sname}") for tx in u_tickers]
        ):
            raise ValueError(f"Contract identifier `{contx}` not in dataframe.")

    ## Apply the slip
    df: pd.DataFrame = _apply_slip(
        df=df,
        slip=slip,
        fids=fids,
    )

    # TODO why pivot it out to a wide format?
    # df_wide = qdf_to_ticker_df(df)
    df_wide = QuantamentalDataFrame(df=df).to_wide()
    return_df = None
    if leverage:
        return_df: pd.DataFrame = _leverage_positions(
            df_wide=df_wide,
            sname=sname,
            fids=fids,
            aum=aum,
            leverage=leverage,
            pname=pname,
        )

    else:
        return_df, pvol, vcv_df = _vol_target_positions(
            df_wide=df_wide,
            sname=sname,
            fids=fids,
            aum=aum,
            vol_target=vol_target,
            rebal_freq=rebal_freq,
            lback_meth=lback_meth,
            est_freqs=est_freqs,
            est_weights=est_weights,
            lback_periods=lback_periods,
            half_life=half_life,
            rstring=rstring,
            pname=pname,
            nan_tolerance=nan_tolerance,
            remove_zeros=remove_zeros,
        )

    # return ticker_df_to_qdf(df=return_df).dropna()
    return_pvol = return_pvol and (locals().get("pvol") is not None)
    return_vcv = return_vcv and (locals().get("vcv_df") is not None)

    # return_df = ticker_df_to_qdf(df=return_df).dropna()
    return_df = QuantamentalDataFrame.from_wide(
        df=return_df, categorical=_initialized_as_categorical
    )
    if return_pvol and return_vcv:
        return (return_df, pvol, vcv_df)
    elif return_pvol:
        return (return_df, pvol)
    elif return_vcv:
        return (return_df, vcv_df)
    else:
        return return_df


if __name__ == "__main__":
    from macrosynergy.management.simulate import make_test_df
    from contract_signals import contract_signals

    cids: List[str] = ["USD", "EUR", "GBP", "AUD", "CAD"]
    xcats: List[str] = ["SIG", "HR"]

    start: str = "2000-01-01"
    end: str = "2020-12-31"

    df: pd.DataFrame = make_test_df(
        cids=cids,
        xcats=xcats,
        start=start,
        end=end,
    )

    df.loc[(df["cid"] == "USD") & (df["xcat"] == "SIG"), "value"] = 1.0
    ctypes = ["FX", "IRS", "CDS"]
    cscales = [1.0, 0.5, 0.1]
    csigns = [1, -1, 1]

    hbasket = ["USD_EQ", "EUR_EQ"]
    hscales = [0.7, 0.3]

    df_cs: pd.DataFrame = contract_signals(
        df=df,
        sig="SIG",
        cids=cids,
        ctypes=ctypes,
        cscales=cscales,
        csigns=csigns,
        hbasket=hbasket,
        hscales=hscales,
        hratios="HR",
    )

    fids: List[str] = [f"{cid}_{ctype}" for cid in cids for ctype in ctypes]

    df_notional: pd.DataFrame = notional_positions(
        df=df_cs,
        fids=fids,
        leverage=1.1,
        sname="STRAT",
    )

    df_xr = make_test_df(
        cids=cids,
        xcats=[f"{_}XR" for _ in ctypes],
        start=start,
        end=end,
    )

    df_notional: pd.DataFrame = notional_positions(
        df=pd.concat([df_cs, df_xr], axis=0),
        fids=fids,
        sname="STRAT",
        vol_target=0.1,
        lback_meth="xma",
        lback_periods=-1,
        half_life=20,
        return_pvol=True,
        return_vcv=True,
    )



from typing import List, Tuple
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
from macrosynergy.management.simulate.simulate_quantamental_data import make_qdf
from macrosynergy.panel.category_relations import CategoryRelations
import textwrap
import seaborn as sns


def multiple_reg_scatter(
    cat_rels: List[CategoryRelations],
    ncol: int = 0,
    nrow: int = 0,
    figsize: Tuple[int, int] = (20, 15),
    title: str = "",
    title_xadj: float = 0.5,
    title_yadj: float = 0.99,
    title_fontsize: int = 20,
    xlab: str = "",
    ylab: str = "",
    label_fontsize: int = 12,
    fit_reg: bool = True,
    reg_ci: int = 95,
    reg_order: int = 1,
    reg_robust: bool = False,
    coef_box: str = None,
    coef_box_size: Tuple[float] = (0.4, 2.5),
    coef_box_font_size: int = 12,
    prob_est: str = "pool",
    separator: int = None,
    single_chart: bool = False,
    subplot_titles: bool = None,
    subplot_title_fontsize: int = 14,
    color_cids: bool = False,
):
    """
    Displays multiple regression scatter plots across categories. The categories are 
    passed as a list of CategoryRelations objects, where the regression calculations 
    take place.

    Parameters
    ----------
    cat_rels : List[CategoryRelations]
        list of CategoryRelations objects.
    ncol : int
        number of columns in the grid. Default is 0, which will be set to the length of
        cat_rels.
    nrow : int
        number of rows in the grid. Default is 0, which will be set to 1.
    figsize : Tuple[float]
        size of the figure. Default is (20, 15).
    title : str
        title of the figure. Default is an empty string.
    xlab : str
        label of the x-axis. Default is an empty string.
    ylab : str
        label of the y-axis. Default is an empty string.
    fit_reg : bool
        if True (default) a linear regression line is fitted to the data.
    reg_ci : int
        confidence interval for the regression line. Default is 95.
    reg_order : int
        order of the regression line. Default is 1.
    reg_robust : bool
        if True (default is False) robust standard errors are used.
    coef_box : str
        if not None, a box with the coefficients of the regression is displayed. Default
        is None.
    coef_box_font_size : int
        font size of the coefficients box. Default is 12. If set to 0 it automatically
        sets the fontsize according to matplotlib.
    prob_est : str
        method to estimate the probability. Default is 'pool'.
    separator : int
        allows categorizing the scatter analysis by integer. This is done by setting it
        to a year [2010, for instance] which will subsequently split the time-period into
        the sample before (not including) that year and from (including) that year.
    single_chart : bool
        if True (default is False) all the data is plotted in a single chart. If False,
        a grid of charts is created.
    subplot_titles : List[str]
        list of titles for each subplot. Default is None.
    color_cids : bool
        if True (default is False) each cross section is given a distinct color in the 
        scatter plot.
    """

    sns.set_theme(style="whitegrid")
    if ncol == 0:
        ncol = len(cat_rels)
    if nrow == 0:
        nrow = 1
    if subplot_titles is not None:
        if len(subplot_titles) != len(cat_rels):
            raise ValueError(
                "The length of subplot_titles must be equal to the length of cat_rels."
            )

    if separator is not None:
        if separator == "cids":
            raise ValueError(
                "Separator 'cids' is not permitted in multiple_reg_scatter. To get a plot across multiple cids, please specify separator as cids inside reg_scatter."
            )
    single_scatter = color_cids
    separator = "cids" if color_cids else separator
    fig, axes = plt.subplots(
        nrows=nrow, ncols=ncol, figsize=figsize, sharex=True, sharey=True
    )
    fig.suptitle(title, x=title_xadj, y=title_yadj, fontsize=title_fontsize)
    fig.supxlabel(xlab, fontsize=label_fontsize)
    fig.supylabel(ylab, fontsize=label_fontsize)

    for i, cat_rel in enumerate(cat_rels):
        row = i // ncol
        col = i % ncol
        if not isinstance(axes, np.ndarray):
            ax = axes
            ax.set_facecolor("white")
        else:
            ax = axes[i] if (ncol == 1 or nrow == 1) else axes[row, col]
            ax.set_facecolor("white")
        if subplot_titles is not None:
            subplot_title = subplot_titles[i]
        else:
            if cat_rel.years is None:
                dates = (
                    cat_rel.df.index.get_level_values("real_date")
                    .to_series()
                    .dt.strftime("%Y-%m-%d")
                )
                subplot_title = (
                    f"{cat_rel.xcats[0]} and {cat_rel.xcats[1]} "
                    f"from {dates.min()} to {dates.max()}"
                )
            else:
                subplot_title = f"{cat_rel.xcats[0]} and {cat_rel.xcats[1]}"

        width = (figsize[0] // ncol) * 6

        wrapped_title = "\n".join(textwrap.wrap(subplot_title, width=width))
        cat_rel.reg_scatter(
            title=wrapped_title,
            labels=False,
            xlab="",
            ylab="",
            fit_reg=fit_reg,
            reg_ci=reg_ci,
            reg_order=reg_order,
            reg_robust=reg_robust,
            coef_box=coef_box,
            coef_box_size=coef_box_size,
            coef_box_font_size=coef_box_font_size,
            prob_est=prob_est,
            single_chart=single_chart,
            separator=separator,
            ax=ax,
            single_scatter=single_scatter,
            title_fontsize=subplot_title_fontsize
        )

    plt.subplots_adjust(top=title_yadj - 0.01)
    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP", "NZD", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2000-01-01", "2020-12-31", 0.1, 1]
    df_cids.loc["CAD"] = ["2001-01-01", "2020-11-30", 0, 1]
    df_cids.loc["BRL"] = ["2001-01-01", "2020-11-30", -0.1, 2]
    df_cids.loc["GBP"] = ["2002-01-01", "2020-11-30", 0, 2]
    df_cids.loc["NZD"] = ["2002-01-01", "2020-09-30", -0.1, 2]
    df_cids.loc["USD"] = ["2003-01-01", "2020-12-31", -0.1, 2]

    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]
    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = [
        "2000-01-01",
        "2020-10-30",
        1,
        2,
        0.95,
        1,
    ]
    df_xcats.loc["GROWTH"] = ["2001-01-01", "2020-10-30", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2001-01-01", "2020-10-30", 1, 2, 0.8, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {"AUD": ["2000-01-01", "2003-12-31"], "GBP": ["2018-01-01", "2100-01-01"]}

    # All AUD GROWTH locations.
    filt1 = (dfd["xcat"] == "GROWTH") & (dfd["cid"] == "AUD")
    filt2 = (dfd["xcat"] == "INFL") & (dfd["cid"] == "NZD")

    # Reduced DataFrame.
    dfdx = dfd[~(filt1 | filt2)].copy()
    dfdx["ERA"]: str = "before 2007"
    dfdx.loc[dfdx["real_date"].dt.year > 2007, "ERA"] = "from 2010"

    cidx = ["AUD", "CAD", "GBP", "USD"]

    cr1 = CategoryRelations(
        dfdx,
        xcats=["CRY", "XR"],
        freq="M",
        lag=1,
        cids=cidx,
        xcat_aggs=["mean", "sum"],
        start="2001-01-01",
        blacklist=black,
        years=None,
    )

    # cr1.reg_scatter(
    #     labels=False,
    #     single_scatter=True,
    #     title="Carry and Return",
    #     xlab="Carry",
    #     ylab="Return",
    #     prob_est="map",
    #     separator="cids"
    # )

    cr2 = CategoryRelations(
        dfdx,
        xcats=["CRY", "GROWTH"],
        # xcat1_chg="diff",
        freq="M",
        lag=1,
        cids=cidx,
        xcat_aggs=["mean", "sum"],
        start="2001-01-01",
        blacklist=black,
        years=None,
    )

    cr3 = CategoryRelations(
        dfdx,
        xcats=["CRY", "INFL"],
        # xcat1_chg="diff",
        freq="M",
        lag=1,
        cids=cidx,
        xcat_aggs=["mean", "sum"],
        start="2001-01-01",
        blacklist=black,
        years=None,
    )

    cr4 = CategoryRelations(
        dfdx,
        xcats=["CRY", "INFL"],
        # xcat1_chg="diff",
        freq="Q",
        lag=1,
        cids=cidx,
        xcat_aggs=["mean", "sum"],
        start="2001-01-01",
        blacklist=black,
        years=None,
    )

    cr5 = CategoryRelations(
        dfdx,
        xcats=["CRY", "INFL"],
        # xcat1_chg="diff",
        freq="Q",
        lag=1,
        cids=cidx,
        xcat_aggs=["mean", "sum"],
        start="2001-01-01",
        blacklist=black,
        years=None,
    )

    cr6 = CategoryRelations(
        dfdx,
        xcats=["CRY", "INFL"],
        # xcat1_chg="diff",
        freq="Q",
        lag=1,
        cids=cidx,
        xcat_aggs=["mean", "sum"],
        start="2001-01-01",
        blacklist=black,
        years=None,
    )

    multiple_reg_scatter(
        [cr1, cr2, cr3, cr4, cr5, cr6],
        title="Growth trend and subsequent sectoral equity returns.",
        xlab="Real technical growth trend",
        ylab="Excess Return",
        ncol=3,
        nrow=2,
        coef_box="upper right",
        color_cids=True,
    )

    multiple_reg_scatter(
        [cr1, cr2, cr3, cr4, cr5, cr6],
        title="Growth trend and subsequent sectoral equity returns.",
        xlab="Real technical growth trend",
        ylab="Excess Return",
        ncol=6,
        nrow=2,
    )



import warnings
import numpy as np
import pandas as pd

from macrosynergy.management.types import QuantamentalDataFrame
from macrosynergy.management.simulate import make_qdf


def impute_panel(
    df: pd.DataFrame,
    cids: list,
    xcats: list,
    threshold: float = 0.5,
    start_date: str = None,
    impute_empty_tickers: bool = False,
) -> pd.DataFrame:
    """
    Imputes missing values for each category in a long-format panel dataset by a cross-
    sectional mean, conditional on the number of available cross-sections at each
    concerned date exceeding a fraction `threshold` of the total number of cross-
    sections.

    Parameters
    ----------
    df : ~pandas.DataFrame
        the long-format panel dataset
    cids : list
        the list of cross sections to be considered in the imputation
    xcats : list
        the list of categories to be imputed
    threshold : float
        the fraction of available cross-sections at each date
    start_date : str
        the starting date for the imputation
    impute_empty_tickers : bool
        boolean flag for whether to impute missing values for empty tickers

    Returns
    -------
    ~pandas.DataFrame
        the imputed long-format panel data with columns


    .. note::
        This class is still **experimental**: the predictions and the API might change
        without any deprecation cycle.
    """

    warnings.warn(
        "This function is deprecated and will be removed very soon, please use the" 
        "panel_imputer class instead.",
        FutureWarning
    )

    # Checks
    if not isinstance(df, pd.DataFrame):
        raise TypeError("The input `df` must be a pandas DataFrame.")
    if not isinstance(xcats, list):
        raise TypeError("The input `xcats` must be a list.")
    for xcat in xcats:
        if not isinstance(xcat, str):
            raise TypeError("The elements of `xcats` must be strings.")
    if not isinstance(threshold, float):
        raise TypeError("The input `threshold` must be a float.")
    if not 0 <= threshold <= 1:
        raise ValueError("The input `threshold` must be between 0 and 1.")
    if not isinstance(start_date, str) and start_date is not None:
        raise TypeError("The input `start_date` must be a string.")
    if not isinstance(impute_empty_tickers, bool):
        raise TypeError("The input `impute_empty_tickers` must be a boolean.")

    if start_date is not None:
        df = df[df["real_date"] >= start_date]

    complete_df = QuantamentalDataFrame(df)
    _as_categorical = complete_df.InitializedAsCategorical
    complete_df = complete_df.set_index(["cid", "real_date", "xcat"])
    complete_df = complete_df
    if impute_empty_tickers:
        cids_series = pd.Series(cids, name="cid")
    else:
        cids_series = complete_df.index.levels[0]
    full_idx = pd.MultiIndex.from_frame(
        pd.concat(
            [
                pd.MultiIndex.from_product(
                    [
                        cids_series,
                        complete_df.index.levels[1],
                        pd.Series([xcat], name="xcat"),
                    ]
                ).to_frame(index=False)
                for xcat in complete_df.index.levels[2].unique()
            ],
            axis=0,
            ignore_index=True,
        ),
    )

    # reindexing to align all the CIDs on the same timeseries of dates
    complete_df = complete_df.reindex(full_idx).reset_index(drop=False)

    # subsetting to keep only the relevant XCATs across CIDs and dates
    incomplete_mask = (complete_df["xcat"].isin(xcats)) & (
        complete_df["cid"].isin(cids)
    )
    incomplete_df = complete_df.loc[incomplete_mask, :]
    # computing the data availability stats
    incomplete_df["mean_val"] = incomplete_df.groupby(["xcat", "real_date"])[
        "value"
    ].transform("mean")
    incomplete_df["tot"] = incomplete_df.groupby(["xcat", "real_date"])[
        "value"
    ].transform("size")
    incomplete_df["avail"] = incomplete_df.groupby(["xcat", "real_date"])[
        "value"
    ].transform("count")

    # Filling CID-specific values only for the appropriate cids and conditional on representative sample
    mask = (incomplete_df["avail"].div(incomplete_df["tot"]) > threshold) & (
        incomplete_df["value"].isna()
    )
    incomplete_df.loc[mask, "value"] = incomplete_df.loc[mask, "value"].fillna(
        incomplete_df.loc[mask, "mean_val"]
    )

    return QuantamentalDataFrame(
        incomplete_df[complete_df.columns].reset_index(drop=True),
        categorical=_as_categorical,
    )

if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP", "NZD", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2000-01-01", "2020-12-31", 0.1, 1]
    df_cids.loc["CAD"] = ["2001-01-01", "2020-11-30", 0, 1]
    # df_cids.loc["BRL"] = ["2001-01-01", "2020-11-30", -0.1, 2]
    df_cids.loc["GBP"] = ["2002-01-01", "2024-12-30", 0, 2]
    df_cids.loc["NZD"] = ["2002-01-01", "2024-12-30", -0.1, 2]
    df_cids.loc["USD"] = ["2003-01-01", "2024-12-31", -0.1, 2]

    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]
    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2024-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-10-30", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2001-01-01", "2020-10-30", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2001-01-01", "2020-10-30", 1, 2, 0.8, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    dfx = impute_panel(dfd, cids + ["BRL"], ["XR"], threshold=0.0, start_date="2000-01-01", impute_empty_tickers=True)

    print(dfx)


"""
Sequential learning over a panel.
"""

import numbers
import warnings
from abc import ABC
from functools import partial

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from joblib import Parallel, delayed
from sklearn.base import ClassifierMixin, RegressorMixin
from sklearn.model_selection import BaseCrossValidator, GridSearchCV, RandomizedSearchCV
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from tqdm.auto import tqdm

from macrosynergy.compat import JOBLIB_RETURN_AS
from macrosynergy.learning.splitters import (
    ExpandingFrequencyPanelSplit,
    ExpandingIncrementPanelSplit,
)
from macrosynergy.management import categories_df
from macrosynergy.management.types.qdf import QuantamentalDataFrame


class BasePanelLearner(ABC):
    def __init__(
        self,
        df,
        xcats,
        cids=None,
        start=None,
        end=None,
        blacklist=None,
        freq="M",
        lag=1,
        xcat_aggs=["last", "sum"],
        generate_labels=None,
        skip_checks=False,
    ):
        """
        Initialize a sequential learning process over a panel.

        Parameters
        ----------
        df : pd.DataFrame
            Standardized daily quantamental dataframe with the four columns: "cid",
            "xcat", "real_date" and "value".
        xcats : list
            List of xcats to be used in the learning process. The last category in the
            list is the dependent variable, and all preceding categories are the
            independent variables in a supervised learning framework.
        cids : list, optional
            Cross-sections to be included. Default is all in the dataframe.
        start : str, optional
            Start date for considered data in subsequent analysis in ISO 8601 format.
            Default is None i.e. the earliest date in the dataframe.
        end : str, optional
            End date for considered data in subsequent analysis in ISO 8601 format.
            Default is None i.e. the latest date in the dataframe.
        blacklist : list, optional
            Blacklisting dictionary specifying date ranges for which cross-sectional
            information should be excluded. The keys are cross-sections and the values
            are tuples of start and end dates in ISO 8601 format. Default is None.
        freq : str, optional
            Frequency of the data. Default is "M" for monthly.
        lag : int, optional
            Number of periods to lag the independent variables. Default is 1.
        xcat_aggs : list, optional
            List of exactly two aggregation methods for downsampling data to the frequency
            specified in the freq parameter. The first parameter pertains to all
            independent variable downsampling, whilst the second corresponds with the
            target category. Default is ["last", "sum"].
        generate_labels : callable, optional
            Function to transform the dependent variable, usually into
            classification labels. Default is None.
        skip_checks : bool, optional
            Whether to skip the initialization checks. Default is False.

        Notes
        -----
        `BasePanelLearner` is an abstract class that provides the basic structure to
        train statistical machine learning models sequentially over a panel. At each
        model retraining date, an optimal model is chosen out a collection of candidate
        models through a hyperparameter optimization process. Following the model
        selection stage, the model is trained on date preceding the retraining date,
        maintaining the point-in-time principle of JPMaQS. Analytics from the selected
        model are stored for interpretability of the process.

        The hyperparameter optimization process revolves around cross-validation [1]_.
        Cross-validation estimates out-of-sample performance metrics. Each candidate
        model is cross-validated over a collection of possible hyperparameters and the
        model with the best score is chosen. When all possible models are directly
        specified and cross-validated, we call this process a 'grid search'. When models
        are sampled from a given collection, we call this process a 'random search'.
        These are the two methods currently supported by `BasePanelLearner`. In a future
        release, we will add support for Bayesian hyperparameter searches [2]_.

        References
        ----------
        .. [1] Brownlee, J. (2023). A Gentle Introduction to k-fold Cross-Validation.
            https://machinelearningmastery.com/k-fold-cross-validation/

        .. [2] Frazier, P.I., 2018. A tutorial on Bayesian optimization.
            https://arxiv.org/abs/1807.02811
        """
        # Checks
        if not skip_checks:
            self._check_init(
                df,
                xcats,
                cids,
                start,
                end,
                blacklist,
                freq,
                lag,
                xcat_aggs,
                generate_labels,
            )
        # Attributes
        self.df = QuantamentalDataFrame(df)
        self.xcats = xcats
        self.cids = cids
        self.start = start
        self.end = end
        self.blacklist = blacklist
        self.freq = freq
        self.lag = lag
        self.xcat_aggs = xcat_aggs
        self.generate_labels = generate_labels

        # Create long-format dataframe
        df_long = (
            categories_df(
                df=self.df,
                xcats=self.xcats,
                cids=self.cids,
                start=self.start,
                end=self.end,
                blacklist=self.blacklist,
                freq=self.freq,
                lag=self.lag,
                xcat_aggs=self.xcat_aggs,
            )
            .dropna()
            .sort_index()
        )

        # Create X and y
        self.X = df_long.iloc[:, :-1]
        self.y = df_long.iloc[:, -1]

        if self.generate_labels is not None:
            self.y = self.y.apply(self.generate_labels)

        # Store necessary index information
        self.index = self.X.index
        self.date_levels = self.index.get_level_values(1)
        self.xs_levels = self.index.get_level_values(0)
        self.unique_date_levels = sorted(self.date_levels.unique())
        self.unique_xs_levels = sorted(self.xs_levels.unique())

        # Create initial dataframe to store model selection data from the learning process
        self.chosen_models = pd.DataFrame(
            columns=[
                "real_date",
                "name",
                "model_type",
                "score",
                "hparams",
                "n_splits_used",
            ]
        ).astype(
            {
                "real_date": "datetime64[ns]",
                "name": "category",
                "model_type": "category",
                "score": "float32",
                "hparams": "object",
                "n_splits_used": "object",
            }
        )

        self.timestamps = []

    def run(
        self,
        name,
        outer_splitter,
        inner_splitters,
        models,
        hyperparameters,
        scorers,
        search_type="grid",
        normalize_fold_results=False,
        cv_summary="mean",
        n_iter=100,
        split_functions=None,
        n_jobs_outer=-1,
        n_jobs_inner=1,
    ):
        """
        Run a learning process over a panel.

        Parameters
        ----------
        name : str
            Category name for the forecasted panel resulting from the learning process.
        outer_splitter : WalkForwardPanelSplit
            Outer splitter for the learning process.
        inner_splitters : dict
            Inner splitters for the learning process.
        models : dict
            Dictionary of model names and compatible `scikit-learn` model objects.
        hyperparameters : dict
            Dictionary of model names and hyperparameter grids.
        scorers : dict
            Dictionary of `scikit-learn` compatible scoring functions.
        search_type : str
            Search type for hyperparameter optimization. Default is "grid".
            Options are "grid", "prior" and "bayes".
        normalize_fold_results : bool
            Whether to normalize the scores across folds before combining them. Default is
            False.
        cv_summary : str or callable
            Summary function for determining cross-validation scores given scores for
            each validation fold. Default is "mean". Can also be "median" or a function
            that takes a list of scores and returns a single value.
        n_iter : int
            Number of iterations for random or bayesian hyperparameter optimization.
        split_functions : dict, optional
            Dictionary of callables for determining the number of cross-validation
            splits to add to the initial number, as a function of the number of iterations
            passed in the sequential learning process. Keys must match with those in the
            `inner_splitters` dictionary. Default is None.
        n_jobs_outer : int, optional
            Number of jobs to run in parallel for the outer loop. Default is -1.
        n_jobs_inner : int, optional
            Number of jobs to run in parallel for the inner loop. Default is 1.

        Returns
        -------
        list
            List of dictionaries containing the results of the learning process.
        """
        # Checks
        self._check_run(
            name=name,
            outer_splitter=outer_splitter,
            inner_splitters=inner_splitters,
            models=models,
            hyperparameters=hyperparameters,
            scorers=scorers,
            search_type=search_type,
            normalize_fold_results=normalize_fold_results,
            cv_summary=cv_summary,
            n_iter=n_iter,
            split_functions=split_functions,
            n_jobs_outer=n_jobs_outer,
            n_jobs_inner=n_jobs_inner,
        )

        # Determine all outer splits and run the learning process in parallel
        train_test_splits = list(outer_splitter.split(self.X, self.y))

        base_splits = self._get_base_splits(inner_splitters)

        # Return list of results
        optim_results = tqdm(
            Parallel(n_jobs=n_jobs_outer, **JOBLIB_RETURN_AS)(
                delayed(self._worker)(
                    name=name,
                    train_idx=train_idx,
                    test_idx=test_idx,
                    inner_splitters=inner_splitters,
                    models=models,
                    hyperparameters=hyperparameters,
                    scorers=scorers,
                    cv_summary=cv_summary,
                    search_type=search_type,
                    normalize_fold_results=normalize_fold_results,
                    n_iter=n_iter,
                    n_splits_add=self._get_n_splits_add(
                        iteration, outer_splitter, split_functions
                    ),
                    n_jobs_inner=n_jobs_inner,
                    base_splits=base_splits,
                )
                for iteration, (train_idx, test_idx) in enumerate(train_test_splits)
            ),
            total=len(train_test_splits),
        )

        return optim_results

    def _worker(
        self,
        name,
        train_idx,
        test_idx,
        inner_splitters,
        models,
        hyperparameters,
        scorers,
        cv_summary,
        search_type,
        normalize_fold_results,
        n_iter,
        n_splits_add,
        n_jobs_inner,
        base_splits,
        timestamp=None,
    ):
        """
        Worker function for parallel processing of the learning process.

        Parameters
        ----------
        name : str
            Category name for the forecasted panel resulting from the learning process.
        train_idx : np.ndarray
            Training indices for the current outer split.
        test_idx : np.ndarray
            Test indices for the current outer split.
        inner_splitters : dict
            Inner splitters for the learning process.
        models : dict
            Compatible `scikit-learn` model objects.
        hyperparameters : dict
            Hyperparameter grids.
        scorers : dict
            Compatible `scikit-learn` scoring functions.
        cv_summary : str or callable
            Summary function to condense cross-validation scores in each fold to a single
            value, against which different hyperparameter choices can be compared.
        search_type : str
            Search type for hyperparameter optimization. Default is "grid".
            Options are "grid", "prior" and "bayes".
        normalize_fold_results : bool
            Whether to normalize the scores across folds before combining them.
        n_iter : int
            Number of iterations for random or bayesian hyperparameter optimization.
        n_splits_add : list, optional
            List of integers to add to the number of splits for each inner splitter.
            Default is None.
        n_jobs_inner : int
            Number of jobs to run in parallel for the inner loop. Default is 1.
        base_splits : dict
            Dictionary of initial number of splits for each inner splitter.
        timestamp : pd.Timestamp, optional
            Date to record predictions and model diagnostics. Default is None. If None,
            the earliest date in the test set is used (which is then adjusted for lag).

        Returns
        -------
        dict
            Dictionary comprising model selection data and predictive analytics.
        """
        # Train-test split
        X_train, X_test = self.X.iloc[train_idx, :], self.X.iloc[test_idx, :]
        y_train, y_test = self.y.iloc[train_idx], self.y.iloc[test_idx]

        # Determine correct timestamps of test set forecasts
        # First get test index information
        test_index = self.index[test_idx]
        test_xs_levels = self.xs_levels[test_idx]
        test_date_levels = self.date_levels[test_idx]
        sorted_test_date_levels = sorted(test_date_levels.unique())

        # Since the features lag behind the targets, the dates need to be adjusted
        # by the lag applied.
        if self.lag != 0:
            locs: np.ndarray = (
                np.searchsorted(
                    self.unique_date_levels, sorted_test_date_levels, side="left"
                )
                - 1
            )
            adj_test_date_levels: pd.DatetimeIndex = pd.DatetimeIndex(
                [self.unique_date_levels[i] if i >= 0 else pd.NaT for i in locs]
            )
            # Now formulate correct index
            date_map = dict(zip(sorted_test_date_levels, adj_test_date_levels))
            mapped_dates = test_date_levels.map(date_map)
            test_index = pd.MultiIndex.from_arrays(
                [test_xs_levels, mapped_dates], names=["cid", "real_date"]
            )
        else:
            adj_test_date_levels = test_date_levels

        if n_splits_add is not None:
            inner_splitters_adj = inner_splitters.copy()
            for splitter_name, _ in inner_splitters_adj.items():
                if hasattr(inner_splitters_adj[splitter_name], "n_splits"):
                    inner_splitters_adj[splitter_name].n_splits = (
                        base_splits[splitter_name] + n_splits_add[splitter_name]
                    )

        else:
            inner_splitters_adj = inner_splitters

        optim_name, optim_model, optim_score, optim_params = self._model_search(
            X_train=X_train,
            y_train=y_train,
            inner_splitters=inner_splitters_adj,
            models=models,
            hyperparameters=hyperparameters,
            scorers=scorers,
            search_type=search_type,
            normalize_fold_results=normalize_fold_results,
            n_iter=n_iter,
            cv_summary=cv_summary,
            n_jobs_inner=n_jobs_inner,
        )

        split_results = self._get_split_results(
            pipeline_name=name,
            optimal_model=optim_model,
            optimal_model_name=optim_name,
            optimal_model_score=optim_score,
            optimal_model_params=optim_params,
            inner_splitters_adj=inner_splitters_adj,
            X_train=X_train,
            y_train=y_train,
            X_test=X_test,
            y_test=y_test,
            timestamp=adj_test_date_levels.min() if timestamp is None else timestamp,
            adjusted_test_index=test_index,
        )

        return split_results

    def _model_search(
        self,
        X_train,
        y_train,
        inner_splitters,
        models,
        hyperparameters,
        scorers,
        search_type,
        normalize_fold_results,
        n_iter,
        cv_summary,
        n_jobs_inner,
    ):
        """
        Determine optimal model based on cross-validation from a given training set.

        Parameters
        ----------
        X_train : pd.DataFrame
            Training data.
        y_train : pd.Series
            Training target.
        inner_splitters : dict
            Inner splitters for the learning process.
        models : dict
            Compatible `scikit-learn` model objects.
        hyperparameters : dict
            Hyperparameter grids.
        scorers : dict
            Compatible `scikit-learn` scoring functions.
        search_type : str
            Search type for hyperparameter optimization. Default is "grid".
        normalize_fold_results : bool
            Whether to normalize the scores across folds before combining them.
        n_iter : int
            Number of iterations for random or bayesian hyperparameter optimization.
        cv_summary : str or callable
            Summary function to condense cross-validation scores in each fold to a single
            value, against which different hyperparameter choices can be compared.
        n_jobs_inner : int
            Number of jobs to run in parallel for the inner loop.

        Returns
        -------
        tuple
            Optimal model name, optimal model, optimal model score and optimal
            model hyperparameters.
        """
        optim_name = None
        optim_model = None
        optim_score = np.float32("-inf")
        optim_params = {}

        cv_splits = []

        for splitter in inner_splitters.values():
            cv_splits.extend(list(splitter.split(X=X_train, y=y_train)))
        # TODO: instead of picking one model, the best hyperparameters could be selected
        # for each model and then "final" prediction would be the average of the individual
        # predictions. This would be a simple ensemble method.
        for model_name, model in models.items():
            # For each model, find the optimal hyperparameters
            if search_type == "grid":
                search_object = GridSearchCV(
                    estimator=model,
                    param_grid=hyperparameters[model_name],
                    scoring=scorers,
                    n_jobs=n_jobs_inner,
                    refit=partial(
                        self._model_selection,
                        cv_summary=cv_summary,
                        scorers=scorers,
                        normalize_fold_results=normalize_fold_results,
                    ),
                    cv=cv_splits,
                )
            elif search_type == "prior":
                search_object = RandomizedSearchCV(
                    estimator=model,
                    param_distributions=hyperparameters[model_name],
                    n_iter=n_iter,
                    scoring=scorers,
                    n_jobs=n_jobs_inner,
                    refit=partial(
                        self._model_selection,
                        cv_summary=cv_summary,
                        scorers=scorers,
                        normalize_fold_results=normalize_fold_results,
                    ),
                    cv=cv_splits,
                )

            try:
                search_object.fit(X_train, y_train)
            except Exception as e:
                warnings.warn(
                    f"Error running a hyperparameter search for {model_name}: {e}",
                    RuntimeWarning,
                )
                continue

            score = self._model_selection(
                search_object.cv_results_,
                cv_summary,
                scorers,
                normalize_fold_results,
                return_index=False,
            )
            if score > optim_score:
                optim_name = model_name
                optim_model = search_object.best_estimator_
                optim_score = score
                optim_params = search_object.best_params_

        return optim_name, optim_model, optim_score, optim_params

    def _model_selection(
        self, cv_results, cv_summary, scorers, normalize_fold_results, return_index=True
    ):
        """
        Select the optimal hyperparameters based on a `scikit-learn` cv_results dataframe.

        Parameters
        ----------
        cv_results : dict
            Cross-validation results dictionary.
        cv_summary : str or callable
            Summary function to condense cross-validation scores in each fold to a single
            value, against which different hyperparameter choices can be compared.
        scorers : dict
            Compatible `scikit-learn` scoring functions.
        normalize_fold_results : bool
            Whether to normalize the scores across folds before combining them.
        return_index : bool
            Whether to return the index of the best estimator or the maximal score itself.
            Default is True.

        Returns
        -------
        int or float
            Either the index of the best estimator or the maximal score itself.
        """
        cv_results = pd.DataFrame(cv_results)
        metric_columns = [
            col
            for col in cv_results.columns
            if col.startswith("split") and "test" in col
        ]
        if normalize_fold_results:
            cv_results[metric_columns] = StandardScaler().fit_transform(
                cv_results[metric_columns]
            )

        # For each metric, summarise the scores across folds for each hyperparameter choice
        # using cv_summary
        for scorer in scorers.keys():
            # Extract the test scores for each fold for that scorer
            scorer_columns = [col for col in metric_columns if scorer in col]
            if cv_summary == "mean":
                cv_results[f"{scorer}_summary"] = cv_results[scorer_columns].mean(
                    axis=1,
                )
            elif cv_summary == "median":
                cv_results[f"{scorer}_summary"] = cv_results[scorer_columns].median(
                    axis=1
                )
            elif cv_summary == "mean-std":
                cv_results[f"{scorer}_summary"] = cv_results[scorer_columns].mean(
                    axis=1
                ) - cv_results[scorer_columns].std(axis=1)
            elif cv_summary == "mean/std":
                cv_results[f"{scorer}_summary"] = cv_results[scorer_columns].mean(
                    axis=1
                ) / cv_results[scorer_columns].std(axis=1)
            else:
                # TODO: handle NAs?
                cv_results[f"{scorer}_summary"] = cv_results[scorer_columns].apply(
                    cv_summary, axis=1
                )

        # Now scale the summary scores for each scorer
        scaler = StandardScaler()

        summary_cols = [f"{scorer}_summary" for scorer in scorers.keys()]

        if len(scorers) > 1:
            scaler = StandardScaler()
            cv_results[summary_cols] = scaler.fit_transform(cv_results[summary_cols])

        # Now average the summary scores for each scorer
        cv_results["final_score"] = cv_results[summary_cols].mean(axis=1)

        # Return index of best estimator
        # TODO: handle case where multiple hyperparameter choices have the same score
        # We currently return the first one
        if return_index:
            return cv_results["final_score"].idxmax()
        else:
            return cv_results["final_score"].max()

    def _get_split_results(
        self,
        pipeline_name,
        optimal_model,
        optimal_model_name,
        optimal_model_score,
        optimal_model_params,
        inner_splitters_adj,
        X_train,
        y_train,
        X_test,
        y_test,
        timestamp,
        adjusted_test_index,
    ):
        """
        Store model selection information and predictive analytics for training set
        (X_train, y_train).

        Parameters
        ----------
        pipeline_name : str
            Name of the sequential optimization pipeline.
        optimal_model : RegressorMixin or ClassifierMixin or Pipeline
            Optimal model selected for the training set.
        optimal_model_name : str
            Name of the optimal model.
        optimal_model_score : float
            Score of the optimal model.
        optimal_model_params : dict
            Hyperparameters of the optimal model.
        inner_splitters_adj : dict
            Inner splitters for the learning process.
        X_train : pd.DataFrame
            Input feature matrix.
        y_train : pd.Series
            Target variable.
        X_test : pd.DataFrame
            Input feature matrix.
        y_test : pd.Series
            Target variable.
        timestamp : pd.Timestamp
            Model retraining date.
        adjusted_test_index : pd.MultiIndex
            Adjusted test index to account for lagged features.

        Returns
        -------
        dict
            Dictionary containing model selection data and predictive analytics.
        """
        split_result = dict()
        model_result = self._store_model_choice_data(
            pipeline_name,
            optimal_model,
            optimal_model_name,
            optimal_model_score,
            optimal_model_params,
            inner_splitters_adj,
            X_train,
            y_train,
            X_test,
            y_test,
            timestamp,
        )

        split_result.update(model_result)

        split_data = self.store_split_data(
            pipeline_name,
            optimal_model,
            optimal_model_name,
            optimal_model_score,
            optimal_model_params,
            inner_splitters_adj,
            X_train,
            y_train,
            X_test,
            y_test,
            timestamp,
            adjusted_test_index,
        )

        split_result.update(split_data)
        return split_result

    def _store_model_choice_data(
        self,
        pipeline_name,
        optimal_model,
        optimal_model_name,
        optimal_model_score,
        optimal_model_params,
        inner_splitters_adj,
        X_train,
        y_train,
        X_test,
        y_test,
        timestamp,
    ):
        """
        Store model selection information for training set (X_train, y_train).

        Parameters
        ----------
        pipeline_name : str
            Name of the sequential optimization pipeline.
        optimal_model : RegressorMixin or ClassifierMixin or Pipeline
            Optimal model selected for the training set.
        optimal_model_name : str
            Name of the optimal model.
        optimal_model_score : float
            Score of the optimal model.
        optimal_model_params : dict
            Hyperparameters of the optimal model.
        inner_splitters_adj : dict
            Inner splitters for the learning process.
        X_train : pd.DataFrame
            Input feature matrix.
        y_train : pd.Series
            Target variable.
        X_test : pd.DataFrame
            Input feature matrix.
        y_test : pd.Series
            Target variable.
        timestamp : pd.Timestamp
            Model retraining date.

        Returns
        -------
        dict
            Dictionary containing model selection data.
        """
        if optimal_model is None:
            warnings.warn(
                f"No model was selected at time {timestamp}",
                RuntimeWarning,
            )
            optimal_model_name = None
            optimal_model_score = np.float32("-inf")
            optimal_model_params = {}

        data = [
            timestamp,
            optimal_model_name,
            optimal_model_score,
            optimal_model_params,
        ]

        n_splits = {
            splitter_name: splitter.n_splits
            for splitter_name, splitter in inner_splitters_adj.items()
        }
        data.append(n_splits)

        return {"model_choice": data}

    def store_split_data(
        self,
        pipeline_name,
        optimal_model,
        optimal_model_name,
        optimal_model_score,
        optimal_model_params,
        inner_splitters_adj,
        X_train,
        y_train,
        X_test,
        y_test,
        timestamp,
        adjusted_test_index,
    ):
        """
        Store predictive analytics for training set (X_train, y_train).

        Parameters
        ----------
        pipeline_name : str
            Name of the sequential optimization pipeline.
        optimal_model : RegressorMixin or ClassifierMixin or Pipeline
            Optimal model selected for the training set.
        optimal_model_name : str
            Name of the optimal model.
        optimal_model_score : float
            Score of the optimal model.
        optimal_model_params : dict
            Hyperparameters of the optimal model.
        inner_splitters_adj : dict
            Inner splitters for the learning process.
        X_train : pd.DataFrame
            Input feature matrix.
        y_train : pd.Series
            Target variable.
        X_test : pd.DataFrame
            Input feature matrix.
        y_test : pd.Series
            Target variable.
        timestamp : pd.Timestamp
            Model retraining date.
        adjusted_test_index : pd.MultiIndex
            Adjusted test index to account for lagged features.

        Returns
        -------
        dict
            Dictionary containing predictive analytics.
        """
        return dict()

    def get_optimal_models(self, name=None):
        """
        Returns the sequences of optimal models for one or more processes.

        Parameters
        ----------
        name : str or list, optional
            Label of sequential optimization process. Default is all stored in the class
            instance.

        Returns
        -------
        pd.DataFrame
            Pandas dataframe of the optimal models and hyperparameters selected at each
            retraining date.
        """
        if name is None:
            return self.chosen_models
        else:
            if isinstance(name, str):
                name = [name]
            elif not isinstance(name, list):
                raise TypeError(
                    "The process name must be a string or a list of strings."
                )

            for n in name:
                if n not in self.chosen_models.name.unique():
                    raise ValueError(
                        f"""The process name '{n}' is not in the list of already-run
                        pipelines. Please check the name carefully. If correct, please run 
                        calculate_predictions() first.
                        """
                    )
            return self.chosen_models[self.chosen_models.name.isin(name)]

    def models_heatmap(
        self,
        name,
        title=None,
        cap=5,
        figsize=(12, 8),
        title_fontsize=None,
        tick_fontsize=None,
    ):
        """
        Visualized optimal models used for signal calculation.

        Parameters
        ----------
        name : str
            Name of the sequential optimization pipeline.
        title : str, optional
            Title of the heatmap. Default is None. This creates a figure title of the form
            "Model Selection Heatmap for {name}".
        cap : int, optional
            Maximum number of models to display. Default (and limit) is 5. The chosen
            models are the 'cap' most frequently occurring in the pipeline.
        figsize : tuple, optional
            Tuple of floats or ints denoting the figure size. Default is (12, 8).
        title_fontsize : int, optional
            Font size for the title. Default is None.
        tick_fontsize : int, optional
            Font size for the ticks. Default is None.

        Notes
        -----
        This method displays the models selected at each date in time over the span
        of the sequential learning process. A binary heatmap is used to visualise
        the model selection process.
        """
        # Checks
        self._checks_models_heatmap(
            name=name,
            title=title,
            cap=cap,
            figsize=figsize,
            title_fontsize=title_fontsize,
            tick_fontsize=tick_fontsize,
        )

        # Get the chosen models for the specified pipeline to visualise selection.
        chosen_models = self.get_optimal_models(name=name).sort_values(by="real_date")
        chosen_models["model_hparam_id"] = chosen_models.apply(
            lambda row: (
                row["model_type"]
                if row["hparams"] == {}
                else f"{row['model_type']}_"
                + "_".join([f"{key}={value}" for key, value in row["hparams"].items()])
            ),
            axis=1,
        )
        chosen_models["real_date"] = chosen_models["real_date"].dt.date
        model_counts = chosen_models.model_hparam_id.value_counts()
        chosen_models = chosen_models[
            chosen_models.model_hparam_id.isin(model_counts.index[:cap])
        ]

        unique_models = chosen_models.model_hparam_id.unique()
        unique_models = sorted(unique_models, key=lambda x: -model_counts[x])
        unique_dates = chosen_models.real_date.unique()

        # Fill in binary matrix denoting the selected model at each time
        binary_matrix = pd.DataFrame(0, index=unique_models, columns=unique_dates)
        for _, row in chosen_models.iterrows():
            model_id = row["model_hparam_id"]
            date = row["real_date"]
            binary_matrix.at[model_id, date] = 1

        # Display the heatmap.
        plt.figure(figsize=figsize)
        if binary_matrix.shape[0] == 1:
            sns.heatmap(binary_matrix, cmap="binary_r", cbar=False)
        else:
            sns.heatmap(binary_matrix, cmap="binary", cbar=False)
        plt.title(title, fontsize=title_fontsize)

        plt.xticks(fontsize=tick_fontsize)  # X-axis tick font size
        plt.yticks(fontsize=tick_fontsize)
        plt.show()

    def _check_init(
        self,
        df,
        xcats,
        cids,
        start,
        end,
        blacklist,
        freq,
        lag,
        xcat_aggs,
        generate_labels,
    ):
        """
        Checks for the constructor.

        Parameters
        ----------
        df : pd.DataFrame
            Long-format dataframe.
        xcats : list
            List of xcats to be used in the learning process.
        cids : list, optional
            List of cids to be used in the learning process. Default is None.
        start : str, optional
            Start date for the learning process. Default is None.
        end : str, optional
            End date for the learning process. Default is None.
        blacklist : dict, optional
            Dictionary of dates to exclude from the learning process. Default is None.
        freq : str
            Frequency of the data. Options are "D", "W", "M", "Q" and "Y".
        lag : int
            Lag to apply to the features. Default is 0.
        xcat_aggs : list
            List of aggregation functions to apply to the independent and
            dependent variables respectively.
        generate_labels : callable, optional
            Function to generate labels for a supervised learning process.
            Default is None.
        """
        # Dataframe checks
        if not isinstance(df, pd.DataFrame):
            raise TypeError("df must be a pandas DataFrame.")
        if not set(["cid", "xcat", "real_date", "value"]).issubset(df.columns):
            raise ValueError(
                "df must have columns 'cid', 'xcat', 'real_date' and 'value'."
            )
        if len(df) < 1:
            raise ValueError("df must not be empty")

        # categories checks
        if not isinstance(xcats, list):
            raise TypeError("xcats must be a list.")
        if len(xcats) < 2:
            raise ValueError("xcats must have at least two elements.")
        if not all(isinstance(xcat, str) for xcat in xcats):
            raise TypeError("All elements in xcats must be strings.")
        difference_xcats = set(xcats) - set(df["xcat"].unique())
        if difference_xcats != set():
            raise ValueError(f"{str(difference_xcats)} not in the dataframe.")

        # cids checks
        if cids is not None:
            if not isinstance(cids, list):
                raise TypeError("cids must be a list.")
            if not all(isinstance(cid, str) for cid in cids):
                raise TypeError("All elements in cids must be strings.")
            difference_cids = set(cids) - set(df["cid"].unique())
            if difference_cids != set():
                raise ValueError(f"{str(difference_cids)} not in the dataframe.")

        # start checks
        if start is not None:
            if not isinstance(start, str):
                raise TypeError("'start' must be a string.")
            try:
                pd.to_datetime(start)
            except ValueError:
                raise ValueError("'start' must be in ISO 8601 format.")
            if pd.to_datetime(start) > pd.to_datetime(df["real_date"]).max():
                raise ValueError("'start' must be before the last date in the panel.")

        # end checks
        if end is not None:
            if not isinstance(end, str):
                raise TypeError("'end' must be a string.")
            try:
                pd.to_datetime(end)
            except ValueError:
                raise ValueError("'end' must be in ISO 8601 format.")
            if pd.to_datetime(end) < pd.to_datetime(df["real_date"]).min():
                raise ValueError("'end' must be after the first date in the panel.")

        if start is not None and end is not None:
            if pd.to_datetime(start) > pd.to_datetime(end):
                raise ValueError("'start' must be before 'end'.")

        # blacklist checks
        if blacklist is not None:
            if not isinstance(blacklist, dict):
                raise TypeError("The blacklist argument must be a dictionary.")
            for key, value in blacklist.items():
                # check keys are strings
                if not isinstance(key, str):
                    raise TypeError(
                        "The keys of the blacklist argument must be strings."
                    )
                # check values of tuples of length two
                if not isinstance(value, tuple):
                    raise TypeError(
                        "The values of the blacklist argument must be tuples."
                    )
                if len(value) != 2:
                    raise ValueError(
                        "The values of the blacklist argument must be tuples of length "
                        "two."
                    )
                # ensure each of the dates in the dictionary are timestamps
                for date in value:
                    if not isinstance(date, pd.Timestamp):
                        raise TypeError(
                            "The values of the blacklist argument must be tuples of "
                            "pandas Timestamps."
                        )

        # freq checks
        if not isinstance(freq, str):
            raise TypeError("freq must be a string.")
        if freq not in ["D", "W", "M", "Q", "Y"]:
            raise ValueError("freq must be one of 'D', 'W', 'M', 'Q' or 'Y'.")

        # lag checks
        if not isinstance(lag, int):
            raise TypeError("lag must be an integer.")
        if lag < 0:
            raise ValueError("lag must be non-negative.")

        # xcat_aggs checks
        if not isinstance(xcat_aggs, list):
            raise TypeError("xcat_aggs must be a list.")
        if not all(isinstance(xcat_agg, str) for xcat_agg in xcat_aggs):
            raise ValueError("All elements in xcat_aggs must be strings.")
        if len(xcat_aggs) != 2:
            raise ValueError("xcat_aggs must have exactly two elements.")

        # generate_labels checks
        if generate_labels is not None:
            if not callable(generate_labels):
                raise TypeError("generate_labels must be a callable.")

    def _check_run(
        self,
        name,
        outer_splitter,
        inner_splitters,
        models,
        hyperparameters,
        scorers,
        normalize_fold_results,
        search_type,
        cv_summary,
        n_iter,
        split_functions,
        n_jobs_outer,
        n_jobs_inner,
    ):
        """
        Input parameter checks for the run method.

        Parameters
        ----------
        name : str
            Name of the sequential optimization pipeline.
        outer_splitter : BasePanelSplit
            Outer splitter for the learning process.
        inner_splitters : dict
            Inner splitters for the learning process.
        models : dict
            Compatible `scikit-learn` model objects.
        hyperparameters : dict
            Hyperparameter grids.
        scorers : dict
            Compatible `scikit-learn` scoring functions.
        normalize_fold_results : bool
            Whether to normalize the scores across folds before combining them.
        search_type : str
            Search type for hyperparameter optimization.
        cv_summary : str or callable
            Summary function to condense cross-validation scores in each fold to a single
            value, against which different hyperparameter choices can be compared.
        n_iter : int
            Number of iterations for random or bayesian hyperparameter optimization.
        split_functions : dict
            Dictionary of functions associated with each inner splitter describing how to
            increase the number of splits as a function of the number of iterations passed.
        n_jobs_outer : int
            Number of jobs to run in parallel for the outer loop.
        n_jobs_inner : int
            Number of jobs to run in parallel for the inner loop.
        """
        # name
        if not isinstance(name, str):
            raise TypeError("name must be a string.")

        # outer splitter
        # TODO: come back and change to WalkForwardPanelSplit
        if outer_splitter:
            if not isinstance(
                outer_splitter, (ExpandingFrequencyPanelSplit, ExpandingIncrementPanelSplit)
            ):
                raise TypeError(
                    "outer_splitter must be an instance of ExpandingFrequencyPanelSplit or ExpandingIncrementPanelSplit."
                )

        # inner splitters
        if not isinstance(inner_splitters, dict):
            raise TypeError("inner splitters should be specified as a dictionary")
        if inner_splitters == {}:
            raise ValueError("The inner splitters dictionary cannot be empty.")

        for names in inner_splitters.keys():
            if not isinstance(names, str):
                raise ValueError(
                    "The keys of the inner splitters dictionary must be strings."
                )
            if not isinstance(inner_splitters[names], BaseCrossValidator):
                raise ValueError(
                    "The values of the inner splitters dictionary must be instances of BaseCrossValidator."
                )

        # models
        if not isinstance(models, dict):
            raise TypeError("The models argument must be a dictionary.")
        if models == {}:
            raise ValueError("The models dictionary cannot be empty.")
        for key in models.keys():
            if not isinstance(key, str):
                raise ValueError("The keys of the models dictionary must be strings.")
            if not isinstance(models[key], (RegressorMixin, ClassifierMixin, Pipeline)):
                raise ValueError(
                    "The values of the models dictionary must be sklearn predictors or "
                    "pipelines."
                )

        # hyperparameters
        if not isinstance(hyperparameters, dict):
            raise TypeError("The hyperparameters argument must be a dictionary.")
        for pipe_name, pipe_params in hyperparameters.items():
            if not isinstance(pipe_name, str):
                raise ValueError(
                    "The keys of the hyperparameters dictionary must be strings."
                )
            if isinstance(pipe_params, dict):
                self._check_hyperparam_grid(search_type, pipe_params)
            elif isinstance(pipe_params, list):
                for param_set in pipe_params:
                    if not isinstance(param_set, dict):
                        raise ValueError(
                            "The values of the hyperparameters dictionary must be dictionaries or lists."
                        )
                    self._check_hyperparam_grid(search_type, param_set)
            else:
                raise ValueError(
                    "The values of the hyperparameters dictionary must be dictionaries or lists."
                )
        # Check that the keys of the hyperparameter grid match those in the models dict
        if sorted(hyperparameters.keys()) != sorted(models.keys()):
            raise ValueError(
                "The keys in the hyperparameter grid must match those in the models "
                "dictionary."
            )

        # scorers
        if not isinstance(scorers, dict):
            raise TypeError("scorers must be a dictionary.")
        if scorers == {}:
            raise ValueError("The scorers dictionary cannot be empty.")
        for key in scorers.keys():
            if not isinstance(key, str):
                raise ValueError("The keys of the scorers dictionary must be strings.")
            if not callable(scorers[key]):
                raise ValueError(
                    "The values of the scorers dictionary must be callable scoring functions."
                )

        # search_type
        if not isinstance(search_type, str):
            raise TypeError("search_type must be a string.")
        if search_type not in ["grid", "prior", "bayes"]:
            raise ValueError("search_type must be one of 'grid', 'prior' or 'bayes'.")
        if search_type == "bayes":
            raise NotImplementedError(
                "Bayesian hyperparameter search is not yet implemented."
            )

        # cv_summary
        if not isinstance(cv_summary, str) and not callable(cv_summary):
            raise TypeError("cv_summary must be a string or a callable.")
        if isinstance(cv_summary, str):
            if cv_summary not in [
                "mean",
                "median",
                "mean-std",
                "mean/std",
            ]:
                raise ValueError(
                    "cv_summary must be one of 'mean', 'median', 'mean-std' or 'mean/std'"
                )
        else:
            try:
                test_summary = cv_summary([1, 2, 3])
            except Exception as e:
                raise TypeError(
                    "cv_summary must be a function that takes a list of scores and returns "
                    "a single value. Check the validity of cv_summary. Error raised when "
                    "testing the function with [1, 2, 3]: {e}"
                )
            if not isinstance(test_summary, numbers.Number) and not isinstance(bool):
                raise TypeError(
                    "cv_summary must be a function that takes a list of scores and returns "
                    "a single value. Check whether the output of cv_summary is a number."
                )

        # n_iter
        if search_type == "prior":
            if not isinstance(n_iter, int):
                raise TypeError("If search_type is 'prior', n_iter must be an integer.")
            if n_iter < 1:
                raise ValueError("The n_iter argument must be greater than zero.")
        elif n_iter is not None and not isinstance(n_iter, int):
            raise ValueError("n_iter must only be used if search_type is 'prior'.")

        # normalize_fold_results
        if not isinstance(normalize_fold_results, bool):
            raise TypeError("normalize_fold_results must be a boolean.")
        if normalize_fold_results:
            if search_type == "grid":
                for model in hyperparameters.keys():
                    num_models = sum(
                        [
                            len(hyperparameters[model][hparam])
                            for hparam in hyperparameters[model].keys()
                        ]
                    )
                    if num_models < 2:
                        raise ValueError(
                            "normalize_fold_results cannot be True if there are less than 2 candidate models. "
                            f"This is the case for the model {model}."
                        )
                    if num_models == 2:
                        warnings.warn(
                            "normalize_fold_results is True but there are only two candidate models for "
                            f"the model {model}. It is recommended for at least three candidate models "
                            "to be available for normalization to be meaningful.",
                            UserWarning,
                        )
            elif search_type == "prior":
                if n_iter < 2:
                    raise ValueError(
                        "normalize_fold_results cannot be True if n_iter is less than 2."
                    )
                if n_iter == 2:
                    warnings.warn(
                        "normalize_fold_results is True but n_iter is 2. It is recommended for n_iter to be "
                        "at least 3 for normalization to be meaningful.",
                        UserWarning,
                    )

        # split_functions
        if split_functions is not None:
            if not isinstance(split_functions, dict):
                raise TypeError("split_functions must be a dictionary.")
            if len(
                set(split_functions.keys()).intersection(set(inner_splitters.keys()))
            ) != len(inner_splitters):
                raise ValueError(
                    "The keys of the split_functions dictionary must match the keys of the inner_splitters dictionary."
                )
            for key in split_functions.keys():
                if not isinstance(key, str):
                    raise ValueError(
                        "The keys of the split_functions dictionary must be strings."
                    )
                if split_functions[key] is not None:
                    if not callable(split_functions[key]):
                        raise ValueError(
                            "The values of the split_functions dictionary must be callables or None."
                        )

        # n_jobs_outer
        if not isinstance(n_jobs_outer, int):
            raise TypeError("n_jobs_outer must be an integer.")
        if n_jobs_outer < 1:
            if n_jobs_outer != -1:
                raise ValueError(
                    "n_jobs_outer must be greater than zero or equal to -1."
                )

        # n_jobs_inner
        if not isinstance(n_jobs_inner, int):
            raise TypeError("n_jobs_inner must be an integer.")
        if n_jobs_inner < 1:
            if n_jobs_inner != -1:
                raise ValueError(
                    "n_jobs_inner must be greater than zero or equal to -1."
                )

    def _check_hyperparam_grid(self, search_type, pipe_params):
        if pipe_params != {}:
            for hparam_key, hparam_values in pipe_params.items():
                if not isinstance(hparam_key, str):
                    raise ValueError(
                        "The keys of the inner hyperparameters dictionaries must be "
                        "strings."
                    )
                if search_type == "grid":
                    if not isinstance(hparam_values, list):
                        raise ValueError(
                            "The values of the inner hyperparameters dictionaries must be "
                            "lists if hparam_type is 'grid'."
                        )
                    if len(hparam_values) == 0:
                        raise ValueError(
                            "The values of the inner hyperparameters dictionaries cannot be "
                            "empty lists."
                        )
                elif search_type == "prior":
                    # hparam_values must either be a list or a scipy.stats distribution
                    # create typeerror
                    if isinstance(hparam_values, list):
                        if len(hparam_values) == 0:
                            raise ValueError(
                                "The values of the inner hyperparameters dictionaries cannot "
                                "be empty lists."
                            )
                    else:
                        if not hasattr(hparam_values, "rvs"):
                            raise ValueError(
                                "Invalid random hyperparameter search dictionary element "
                                f"for hyperparameter {hparam_key}. The dictionary values "
                                "must be scipy.stats distributions."
                            )

    def _checks_models_heatmap(
        self,
        name,
        title=None,
        cap=5,
        figsize=(12, 8),
        title_fontsize=None,
        tick_fontsize=None,
    ):
        """
        Checks for the models_heatmap method.

        Parameters
        ----------
        name : str
            Name of the sequential optimization pipeline.
        title : str, optional
            Title of the heatmap. Default has the form "Model Selection Heatmap for {name}".
        cap : int, optional
            Maximum number of models to display. Default is 5, with limit of 10.
            The chosen models are the 'cap' most frequently occurring in the process.
        figsize : tuple, optional
            Tuple of floats or ints denoting the figure size. Default is (12, 8).
        title_fontsize : int, optional
            Font size for the title. Default is None.
        tick_fontsize : int, optional
            Font size for the ticks. Default is None.
        """
        if not isinstance(name, str):
            raise TypeError("The pipeline name must be a string.")
        if name not in self.chosen_models.name.unique():
            raise ValueError(
                f"""The pipeline name {name} is not in the list of already-calculated 
                pipelines. Please check the pipeline name carefully. If correct, please 
                run calculate_predictions() first.
                """
            )
        if not isinstance(cap, int):
            raise TypeError("The cap must be an integer.")
        if cap < 0:
            raise ValueError("The cap must be greater than zero.")
        if cap > 10:
            raise ValueError("The cap must be 10 or lower.")

        if title is None:
            title = f"Model Selection Heatmap for {name}"
        if not isinstance(title, str):
            raise TypeError("The figure title must be a string.")

        if not isinstance(figsize, tuple):
            raise TypeError("The figsize argument must be a tuple.")
        if len(figsize) != 2:
            raise ValueError("The figsize argument must be a tuple of length 2.")
        for element in figsize:
            if not isinstance(element, (int, float)):
                raise TypeError(
                    "The elements of the figsize tuple must be floats or ints."
                )

        if title_fontsize is not None:
            if not isinstance(title_fontsize, int):
                raise TypeError("The title_fontsize argument must be an integer.")
            if title_fontsize < 0:
                raise ValueError("The title_fontsize argument must be non-negative.")

        if tick_fontsize is not None:
            if not isinstance(tick_fontsize, int):
                raise TypeError("The tick_fontsize argument must be an integer.")
            if tick_fontsize < 0:
                raise ValueError("The tick_fontsize argument must be non-negative.")

    def _get_base_splits(self, inner_splitters):
        """Get the initial number of splits for each splitter."""
        base_splits = {}
        for splitter_name, splitter in inner_splitters.items():
            if hasattr(splitter, "n_splits"):
                base_splits[splitter_name] = splitter.n_splits
        return base_splits

    def _get_n_splits_add(self, iteration, outer_splitters, split_functions) -> dict:
        """Call split functions to determine the number of splits to add to each splitter."""
        if split_functions is not None:
            return {
                splitter_name: (
                    int(np.ceil(split_function(iteration * outer_splitters.test_size)))
                    if split_function is not None
                    else 0
                )
                for splitter_name, split_function in split_functions.items()
            }
        else:
            return None

    def _remove_results(self, conditions):
        """
        Remove rows from results DataFrames based on conditions.

        Parameters
        ----------
        conditions : list of tuples
            List of tuples containing the attribute name, column name and value to filter on.
        """
        for attr, column, value in conditions:
            df = getattr(self, attr)
            if value in df[column].unique():
                setattr(self, attr, df[df[column] != value])



"""
Implementation of adjust_weights.
"""

import numpy as np
import pandas as pd
from typing import List, Tuple, Callable, Dict, Any, Optional
import warnings
from numbers import Number
from macrosynergy.management.utils import reduce_df, get_cid
from macrosynergy.management.simulate import make_test_df
from macrosynergy.management.types import QuantamentalDataFrame
from macrosynergy.compat import PD_NEW_MAP

AVAILABLE_METHODS: List[str] = ["generic", "lincomb"]


def check_missing_cids_xcats(weights, adj_zns, cids, r_xcats, r_cids):
    """
    Checks if there are missing cids or xcats in the input DataFrame.
    """
    missing_xcats = list(set([weights, adj_zns]) - set(r_xcats))
    if missing_xcats:
        raise ValueError(f"Missing xcats: {missing_xcats}")

    missing_cids = list(set(cids) - set(r_cids))
    if missing_cids:
        raise ValueError(f"Missing cids: {missing_cids}")


def check_types(
    weights: str,
    adj_zns: str,
    method: str,
    adj_func: Callable,
    params: Dict[str, Any],
    cids: List[str],
    start: Optional[str] = None,
    end: Optional[str] = None,
):
    """
    Type checking for the input variables of adjust_weights.
    """
    for _var, _name, _type in [
        (weights, "weights", str),
        (adj_zns, "adj_zns", str),
        (method, "method", str),
        (adj_func, "adj_func", (Callable, type(None))),
        (params, "param", dict),
        (cids, "cids", (list, type(None))),
        (start, "start", (str, type(None))),
        (end, "end", (str, type(None))),
    ]:
        if not isinstance(_var, _type):
            raise TypeError(f"{_name} must be a {_type}, not {type(_var)}")

    if cids is not None and (
        not all(isinstance(cid, str) for cid in cids) or len(cids) == 0
    ):
        raise TypeError("`cids` must be a None(default) or a non-empty list of strings")

    if method not in AVAILABLE_METHODS:
        raise ValueError(
            f"Method {method} not available. Available methods: {AVAILABLE_METHODS}"
        )
    if method == "generic":
        if adj_func is None:
            raise ValueError("`adj_func` must be provided when method='generic'")


def lincomb_backend(
    df_adj_zns_wide: pd.DataFrame,
    df_weights_wide: pd.DataFrame,
    coeff_new: float,
    min_score: Optional[float] = None,
) -> pd.DataFrame:
    """
    Linear combination of the parameters.

    Parameters
    ----------
    df_adj_zns_wide : pd.DataFrame
        DataFrame with adjustment factors in wide format.
    df_weights_wide : pd.DataFrame
        DataFrame with weights in wide format.
    coeff_new : float
        Coefficient (between 0 and 1) for the new weights. 1 means the result consists
        entirely of the new weights, 0 means the result consists entirely of the old
        weights.
    min_score : float, optional
        Minimum score for the adjustment factors. Default is None, where it is set to the
        minimum score discovered in the panel of `df_adj_zns_wide`.
    """

    assert set(df_weights_wide.columns) == set(df_adj_zns_wide.columns)
    assert set(df_weights_wide.index) == set(df_adj_zns_wide.index)

    if min_score is None:
        warnings.warn(
            "`min_score` not provided. Defaulting to minimum value from `df_adj_zns_wide`."
        )
        min_score = df_adj_zns_wide.min().min()

    err_str = "Parameter `coeff_new` must be provided as a floating point number between 0 and 1."
    if not isinstance(coeff_new, Number) or (
        isinstance(coeff_new, Number) and not 0 <= coeff_new <= 1
    ):
        raise ValueError(err_str)

    # Algorithm:
    # new_weight_basis[i, t] = max(adj_zns[i, t] - min_score, 0)
    # new_weight[i, t] = new_weight_basis[i, t] / sum(new_weight_basis[t])
    # output_raw_weight[i, t] = (1 - coeff_new) * old_weight[i, t] + coeff_new * new_weight[i, t]
    # output_weight[i, t] = output_raw_weight[i, t] / sum(output_raw_weight[i, t]))
    # where `i` is the cross-section and `t` is the date

    nwb = df_adj_zns_wide - min_score
    nwb[nwb < 0] = 0
    nw = nwb.div(nwb.sum(axis="columns"), axis="index")
    orw = (1 - coeff_new) * df_weights_wide + coeff_new * nw
    ow = orw.div(orw.sum(axis="columns"), axis="index")

    return ow


def generic_weights_backend(
    df_weights_wide: pd.DataFrame,
    df_adj_zns_wide: pd.DataFrame,
    adj_func: Callable,
    params: Dict[str, Any] = {},
) -> pd.DataFrame:
    """
    Backend function for adjust_weights. Applies the `method` function to the weights and
    multiplies the result by the adjustment factors, and by the parameter `param`.
    Expects the input DataFrames to be in wide format, with the same columns AND index
    (see macrosynergy.panel.adjust_weights.split_weights_adj_zns).

    Parameters
    ----------
    df_weights_wide : pd.DataFrame
        DataFrame with weights in wide format.

    df_adj_zns_wide : pd.DataFrame
        DataFrame with adjustment factors in wide format.

    method : Callable
        Function that will be applied to the weights to adjust them.

    params : Dict[str, Any], optional
        Parameters to be passed to the method function. Default is {}.

    Returns
    -------
    pd.DataFrame
        DataFrame with the adjusted weights.
    """

    assert set(df_weights_wide.columns) == set(df_adj_zns_wide.columns)
    assert set(df_weights_wide.index) == set(df_adj_zns_wide.index)

    if PD_NEW_MAP:
        dfw_result = df_weights_wide * df_adj_zns_wide.map(adj_func, **params)
    else:
        dfw_result = df_weights_wide * df_adj_zns_wide.applymap(adj_func, **params)

    return dfw_result


def split_weights_adj_zns(
    df: QuantamentalDataFrame, weights: str, adj_zns: str
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Splits the input DataFrame into two DataFrames, one containing the weights and the
    other containing the adjustment factors.

    Parameters
    ----------
    df : QuantamentalDataFrame
        DataFrame containing the weights and adjustment factors.

    weights : str
        Name of the xcat to be used as weights.

    adj_zns : str
        Name of the z-n score xcat to be used as adjustment factors.

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame]
        Tuple containing two wide DataFrames (one for weights and one for adjustment
        factors), with one column per cid.
    """

    df_weights_wide = QuantamentalDataFrame(df.loc[df["xcat"] == weights]).to_wide()
    df_adj_zns_wide = QuantamentalDataFrame(df.loc[df["xcat"] == adj_zns]).to_wide()

    # cannot tolerate negative weights
    if any(df_weights_wide[~df_weights_wide.isna()].lt(0).any()):
        na_frame = QuantamentalDataFrame.from_wide(
            df_weights_wide[
                df_weights_wide[~df_weights_wide.isna()].lt(0).any(axis="columns")
            ]
        )

        na_frame = na_frame[na_frame["value"] < 0]
        raise ValueError(
            f"Negative weights found in the dataframe. Please check the following data:\n{na_frame}"
        )

    combined_index = df_weights_wide.index.union(df_adj_zns_wide.index)
    df_weights_wide = df_weights_wide.reindex(combined_index)
    df_adj_zns_wide = df_adj_zns_wide.reindex(combined_index)

    df_weights_wide.columns = get_cid(df_weights_wide.columns)
    df_adj_zns_wide.columns = get_cid(df_adj_zns_wide.columns)

    zns_missing_in_weights = set(df_adj_zns_wide.columns) - set(df_weights_wide.columns)
    weights_missing_in_zns = set(df_weights_wide.columns) - set(df_adj_zns_wide.columns)
    zns_missing_in_weights = [f"{c}_{adj_zns}" for c in zns_missing_in_weights]
    weights_missing_in_zns = [f"{c}_{weights}" for c in weights_missing_in_zns]
    all_missing = zns_missing_in_weights + weights_missing_in_zns
    if all_missing:
        raise ValueError(f"Missing tickers: {all_missing}")

    # get the corresponding rows in zns
    nan_zns_rows = df_adj_zns_wide.isna().all(axis="columns")
    all_zero_zns_rows = (df_adj_zns_wide.fillna(0) == 0).all(axis="columns")
    missing_zns_dates = df_adj_zns_wide.index[nan_zns_rows | all_zero_zns_rows]

    nan_weights_rows = df_weights_wide.isna().all(axis="columns")
    all_zero_weights_rows = (df_weights_wide.fillna(0) == 0).all(axis="columns")
    missing_weights_dates = df_weights_wide.index[
        nan_weights_rows | all_zero_weights_rows
    ]

    # if zn is missing, but weight is not missing, fill zn with 1
    missing_zns_dates = sorted(set(missing_zns_dates) - set(missing_weights_dates))
    if len(missing_zns_dates) > 0:
        estr = "Missing ZNs data (will be filled with 1 to preserve weights):"
        warnings.warn(f"{estr} {missing_zns_dates}")

        # replace missing zns data with standard weights
        df_adj_zns_wide.loc[missing_zns_dates] = 1

    return df_weights_wide, df_adj_zns_wide


def normalize_weights(
    out_weights: pd.DataFrame, normalize_to_pct: bool = False
) -> pd.DataFrame:
    """
    Output weights are normalized by dividing each row by the sum of the row. Function exists to
    allow easy modification of normalization method.

    Parameters
    ----------
    out_weights : pd.DataFrame
        DataFrame with weights in wide format. (one column per cid)

    normalize_to_pct : bool, optional
        If True, the resulting weights will be scaled to 100%. Default is False.

    Returns
    -------
    pd.DataFrame
        DataFrame with normalized weights (sum of each row is 1).
    """
    out_weights = out_weights.div(out_weights.sum(axis="columns"), axis="index")

    norm_rows = out_weights.sum(axis="columns").apply(lambda x: np.isclose(x, 1))
    all_nan_rows = out_weights.index[out_weights.isnull().all(axis="columns")]

    # assert that all rows sum to 1 or are all NaN
    if not norm_rows.all() and all_nan_rows.size == 0:
        raise Exception("Normalization failed; weights do not sum to 1")

    if normalize_to_pct:
        out_weights = out_weights * 100

    return out_weights


def adjust_weights(
    df: QuantamentalDataFrame,
    weights_xcat: str,
    adj_zns_xcat: str,
    method: str = "generic",
    adj_func: Callable = None,
    params: Dict[str, Any] = {},
    cids: List[str] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    blacklist: Dict[str, Any] = None,
    normalize: bool = True,
    normalize_to_pct: bool = False,
    adj_name: str = "ADJWGT",
):
    """
    Adjusts the weights of a given xcat by a given adjustment xcat using a given method.
    The resulting weights will be scaled to sum to 100% for each date.

    Parameters
    ----------
    df : QuantamentalDataFrame
        QuantamentalDataFrame with weights and adjustment categories for all cross-sections.
    weights_xcat : str
        Name of the category containing the weights.
    adj_zns_xcat : str
        Name of the category containing the adjustment factors.
    method : Callable
        One of the available methods for adjusting weights. Default is "generic".
        See notes for available methods.
    adj_func : Callable, optional
        Function to be used for the adjustment when method is "generic". This function will
        be applied to the weights and multiplied by the adjustment factors. Default is None.
    params : Dict[str, Any], optional
        Parameters to be passed to the method function. Default is {}.
    cids : List[str], optional
        List of cross-sections to adjust. If None, all cross-sections will be adjusted. Default is None.
    start : str, optional
        Start date for the adjustment as YYYY-MM-DD. Default is None.
    end : str, optional
        End date for the adjustment as YYYY-MM-DD. Default is None.
    blacklist : Dict[str, Any], optional
        Blacklist dictionary passed to the reduce_df function. Default is None.
        See :meth:`macrosynergy.management.utils.df_utils.reduce_df` for more details.
    normalize : bool, optional
        If True, the resulting weights will be normalized to sum to one for each date for
        the entire list of cross-sections. Default is True.
    normalize_to_pct : bool, optional
        If True, the resulting weights will be scaled to 100%. Default is False.
        This only applies if `normalize` is True.
    adj_name : str, optional
        Name of the resulting xcat. Default is "ADJWGT".


    Returns
    -------
    QuantamentalDataFrame
        DataFrame with the adjusted weights.

    Notes
    -----
    Available methods:
    - "generic": Applies the method function to the weights and multiplies the result by the
        adjustment factors. The `method` function's signature must match:
        `method(weight: float, **params) -> float`.

    - "lincomb": Linear combination of the parameters. The method function must accept a single
        argument (the weight) and return a single value (the adjusted weight). The parameters
        `min_score` (minimum score for the adjustment factors) and `coeff_new` (coefficient for
        the new weights) must be provided in the `params` dictionary. See
        macrosynergy.panel.adjust_weights.lincomb_backend for more details.

    Examples
    --------

    >>> df = make_test_df(xcats=["weights", "adj_zns"], cids=["cid1", "cid2", "cid3"])

    >>>

    """

    if not isinstance(df, QuantamentalDataFrame):
        raise TypeError("df must be a QuantamentalDataFrame")

    df: QuantamentalDataFrame = QuantamentalDataFrame(df)
    result_as_categorical: bool = df.InitializedAsCategorical

    check_types(
        weights=weights_xcat,
        adj_zns=adj_zns_xcat,
        method=method,
        adj_func=adj_func,
        params=params,
        cids=cids,
        start=start,
        end=end,
    )

    df, r_xcats, r_cids = reduce_df(
        df,
        cids=cids,
        xcats=[weights_xcat, adj_zns_xcat],
        start=start,
        end=end,
        blacklist=blacklist,
        intersect=True,
        out_all=True,
    )
    if cids is None:
        cids = df["cid"].unique().tolist()

    check_missing_cids_xcats(weights_xcat, adj_zns_xcat, cids, r_xcats, r_cids)

    df_weights_wide, df_adj_zns_wide = split_weights_adj_zns(
        df, weights_xcat, adj_zns_xcat
    )

    # no need to normalize weights before applying the adjustment

    if method == "lincomb":
        dfw_result = lincomb_backend(
            df_adj_zns_wide=df_adj_zns_wide,
            df_weights_wide=df_weights_wide,
            coeff_new=params.get("coeff_new", None),
            min_score=params.get("min_score", None),
        )

    elif method == "generic":
        dfw_result = generic_weights_backend(
            df_weights_wide=df_weights_wide,
            df_adj_zns_wide=df_adj_zns_wide,
            adj_func=adj_func,
            params=params,
        )
    else:
        # this condition is covered in a check above
        raise ValueError(f"Method {method} not available.")  # pragma: no cover

    all_nan_rows = dfw_result.index[dfw_result.isnull().all(axis="columns")]
    if all_nan_rows.size > 0:
        err = "The following dates have no data after applying the adjustment, and will be dropped:"
        warnings.warn(f"{err} {all_nan_rows}")
        dfw_result = dfw_result.dropna(how="all", axis="rows")
    if normalize:
        # normalize and scale to 100%
        dfw_result = normalize_weights(dfw_result, normalize_to_pct)

    if dfw_result.isna().all().all():
        raise ValueError(
            "The resulting DataFrame is empty. Please check the input data,"
            " the method function, and it's parameters."
        )

    dfw_result.columns += f"_{adj_name}"
    qdf = QuantamentalDataFrame.from_wide(dfw_result, categorical=result_as_categorical)
    qdf = qdf.dropna(how="any", axis=0).reset_index(drop=True)
    return qdf


if __name__ == "__main__":
    df = make_test_df(xcats=["weights", "adj_zns"], cids=["cid1", "cid2", "cid3"])
    dfb = make_test_df(xcats=["some_xcat", "other_xcat"], cids=["cid1", "cid2", "cid4"])

    # nan_mask = np.random.rand(len(df)) < 0.01
    # df.loc[nan_mask, "value"] = np.nan
    # nan_mask = np.random.rand(len(df)) < 0.1
    # df.loc[nan_mask, "value"] *= -1

    df = pd.concat([df, dfb], axis=0)

    # Using the lincomb method

    df_res = adjust_weights(
        df=df,
        weights_xcat="weights",
        adj_zns_xcat="adj_zns",
        method="lincomb",
        params={"min_score": None, "coeff_new": 0.5},
    )

    assert np.allclose(df_res.groupby("real_date")["value"].sum(), 1)

    # Using the generic method

    def sigmoid(x, amplitude=1.0, steepness=1.0, midpoint=0.0):
        """Sigmoid function with parameters for amplitude, steepness, and midpoint."""
        return amplitude / (1 + np.exp(-steepness * (x - midpoint)))

    params = {"amplitude": 1, "steepness": 4, "midpoint": 1}

    df_res = adjust_weights(
        df=df,
        weights_xcat="weights",
        adj_zns_xcat="adj_zns",
        method="generic",
        adj_func=sigmoid,
        params=params,
    )

    assert np.allclose(df_res.groupby("real_date")["value"].sum(), 1)

    print(df_res)



"""
Run Granger Causality Test on a standardized quantamental dataframe.  **NOTE: This
module is under development, and is not yet ready for production use.**
"""

import warnings
from typing import Any, Dict, List, Optional, Union

import pandas as pd
import statsmodels
from packaging import version
from statsmodels.tsa.stattools import grangercausalitytests

from macrosynergy.management.simulate import make_test_df
from macrosynergy.management.types import QuantamentalDataFrame
from macrosynergy.management.utils import (
    downsample_df_on_real_date,
    is_valid_iso_date,
    qdf_to_ticker_df,
    reduce_df_by_ticker,
)

import logging

logger = logging.getLogger(__name__)


def granger_causality_test(
    df: pd.DataFrame,
    tickers: Optional[List[str]] = None,
    cids: Optional[Union[str, List[str]]] = None,
    xcats: Optional[Union[str, List[str]]] = None,
    max_lag: Union[int, List[int]] = 4,
    add_constant: bool = False,
    freq: str = "M",
    agg: str = "mean",
    start: Optional[str] = None,
    end: Optional[str] = None,
    metric: str = "value",
) -> Dict[Any, Any]:
    """
    Run Granger Causality Test on a standardized quantamental dataframe. Since the Granger
    Causality Test is a pairwise test, only two tickers are permitted. Tickers may be
    specified directly using the `tickers` parameter, or formed by providing one
    cross-section (`cid`) and two extended categories (`xcats`), or two cross-sections
    (`cids`) and one extended category (`xcat`). Tickers are constructed from the
    specified `cids` and `xcats` in the provided order. The order of the tickers is
    preserved, with the first ticker being tested to determine whether it Granger-causes
    the second. The test evaluates whether the time series of the first ticker
    Granger-causes the time series of the second ticker.

    Parameters
    ----------
    df : pd.DataFrame
        A standardized quantamental dataframe.
    tickers : List[str]
        A list of tickers to run the test on. A maximum of two tickers can be specified.
    cids : Union[str, List[str]]
        One or two cross-sections to run the test on. If two cross-sections are specified,
        then only one extended category can be specified. If one cross-section is
        specified, then two categories must be specified.
    xcats : Union[str, List[str]]
        One or two extended categories to run the test on. If two categories are specified,
        then only one cross-sections can be specified. If one category is specified, then
        two cross-sections must be specified.
    max_lag : Union[int, List[int]]
        If `max_lag` is an integer, then the function computes the test for all lags up
        to `max_lag`. If `max_lag` is a list of integers, then the function computes the
        test only for lags specified in the list.
    add_constant : bool
        Whether to add a constant to the regression.
    freq : str
        The frequency to downsample the data to. Must be one of "D", "W", "M", "Q", "A".
        Default is "M".
    agg : str
        The aggregation method to use when downsampling the data. Must be one of "mean"
        (default), "median", "min", "max", "first" or "last".
    start : str
        The start date of the data. Must be a valid ISO date. If not specified, the
        earliest date in `df` is used.
    end : str
        The end date of the data. Must be a valid ISO date. If not specified, the latest
        date in `df` is used.
    metric : str
        The metric to run the test on. Must be a column in `df`. Default is "value".

    Raises
    ------
    TypeError
        If any of the inputs are of the wrong type.
    ValueError
        If any of the input values are invalid.

    Returns
    -------
    Dict[Any, Any]
        A dictionary containing the results of the Granger Causality Test. The keys are
        the lags and the values are the results of the test.
    """

    ## Check inputs

    _type_checks(
        df=df,
        tickers=tickers,
        cids=cids,
        xcats=xcats,
        max_lag=max_lag,
        add_constant=add_constant,
        start=start,
        end=end,
        freq=freq,
        agg=agg,
        metric=metric,
    )
    ## value checks for `freq` and `agg` are implicitly checked in downstream functions

    ## Copy df to prevent side effects
    df: QuantamentalDataFrame = QuantamentalDataFrame(df)

    ## Construct tickers from the `cids` and `xcats` if `tickers` is not specified
    tickers: List[str] = _get_tickers(tickers=tickers, cids=cids, xcats=xcats)

    ## Reduce df
    df: QuantamentalDataFrame = reduce_df_by_ticker(
        df=df, ticks=tickers, start=start, end=end
    )

    # Downsample df
    freq = freq.upper()
    agg = agg.lower()
    df = downsample_df_on_real_date(
        df=df, groupby_columns=["cid", "xcat"], freq=freq, agg=agg
    )

    # Pivot df
    df_wide: pd.DataFrame = QuantamentalDataFrame(df).to_wide(value_column=metric)

    # there must only be two columns in df_wide
    assert len(df_wide.columns) == 2, "df_wide must have only two columns"

    logger.info(
        "Running Granger Causality Test: Testing whether %s Granger causes %s",
        df_wide.columns[0],
        df_wide.columns[1],
    )
    # NOTE: Since no NANs are allowed in the input data, we must drop them here
    # This may yield unexpected/unreliable results for tickers with large periods of
    # missing data

    # drop any rows with NANs
    df_wide = df_wide.dropna(how="any", axis=0)
    if df_wide.empty:
        raise ValueError(
            "The input data contains only NANs. "
            "Please check the input data for missing values or "
            "consider using a different downsampling frequency/date range."
        )

    gct: Dict[Any, Any] = _granger_causality_backend(
        data=df_wide,
        max_lag=max_lag,
    )

    return gct


def _statsmodels_compatibility_wrapper(
    x: Any = None, maxlag: Any = None, addconst: Any = None, verbose: Any = None
) -> Any:
    """
    Wrapper function to handle compatibility issues with different versions of
    statsmodels.
    """
    if version.parse(statsmodels.__version__) < version.parse("0.15.0"):
        with warnings.catch_warnings():
            warnings.simplefilter(action="ignore", category=FutureWarning)
            return grangercausalitytests(x, maxlag, addconst, False)

    else:
        return grangercausalitytests(x, maxlag, addconst)


def _granger_causality_backend(
    data: pd.DataFrame, max_lag: Union[int, List[int]], add_constant: bool = True
) -> Dict[Any, Any]:
    assert len(data.columns) == 2, "`data` must have only two columns"
    assert (
        isinstance(max_lag, int)
        or isinstance(max_lag, list)
        and all(isinstance(l, int) for l in max_lag)
        and len(max_lag) > 0
    ), "`max_lag` must be an integer or a list of integers"
    assert isinstance(add_constant, bool), "`add_constant` must be a boolean"

    arguments: Dict[str, Any] = dict(
        x=data,
        maxlag=max_lag,
        addconst=add_constant,
    )

    return _statsmodels_compatibility_wrapper(**arguments)


def _get_tickers(
    tickers: Optional[List[str]] = None,
    cids: Optional[List[str]] = None,
    xcats: Optional[List[str]] = None,
) -> List[str]:
    """
    Simply returns the tickers if they are specified. If they are not specified, then
    the function forms the list of tickers from the `cids` and `xcats` parameters such
    that the order of the formed tickers is preserved.

    Parameters
    ----------
    tickers : List[str]
        A list of tickers.
    cids : Union[str, List[str]]
        One or two cross-sections.
    xcats : Union[str, List[str]]
        One or two extended categories.
    """

    if tickers is not None:
        return tickers
    else:
        if isinstance(cids, str):
            cids: List[str] = [cids]
        if isinstance(xcats, str):
            xcats: List[str] = [xcats]
        return [f"{c}_{x}" for c in cids for x in xcats]


def _type_checks(
    df: pd.DataFrame,
    tickers: Optional[List[str]],
    cids: Optional[List[str]],
    xcats: Optional[List[str]],
    max_lag: Union[int, List[int]],
    add_constant: bool,
    start: Optional[str],
    end: Optional[str],
    freq: str,
    agg: str,
    metric: str,
) -> bool:
    """
    Does type checks on the inputs to `granger_causality_test`. All inputs are checked
    for type and value errors.

    Raises
    ------
    TypeError
        If any of the inputs are of the wrong type.
    ValueError
        If any of the input values are invalid.

    Returns
    -------
    bool
        True if all type checks pass.
    """

    if not isinstance(df, QuantamentalDataFrame):
        raise TypeError("df must be a standardized quantamental dataframe")

    if not isinstance(metric, str):
        raise TypeError("`metric` must be a string")

    err_msg: str = f"`metric` '{metric}' not found in `df`"
    if metric not in df.columns:
        raise ValueError(err_msg)

    if not isinstance(max_lag, (int, list)):
        raise TypeError("`max_lag` must be an integer or a list of integers")
    elif isinstance(max_lag, list) and not all(isinstance(l, int) for l in max_lag):
        raise TypeError("`max_lag` must be an integer or a list of integers")

    for dt, nm in zip([start, end], ["start", "end"]):
        if dt is not None and not is_valid_iso_date(dt):
            raise ValueError(f"{nm} must be a valid ISO date")

    if isinstance(cids, str):
        cids: List[str] = [cids]
    if isinstance(xcats, str):
        xcats: List[str] = [xcats]

    if bool(cids) ^ bool(xcats):
        raise ValueError("`cids` and `xcats` must be specified together")

    bcidxcats: bool = bool(cids) and bool(xcats)

    if bool(tickers) and (bcidxcats):
        raise ValueError(
            "`tickers` cannot be specified if `cids` & `xcats` are specified"
        )

    found_tickers: List[str] = QuantamentalDataFrame(df).list_tickers()

    if bool(tickers):
        # check if there are only two
        if len(set(tickers)) != 2:
            raise ValueError("Only two tickers can be specified in `tickers`")

        if not all(isinstance(t, str) for t in tickers):
            raise TypeError("`tickers` must be a list of strings")

        if not set(tickers).issubset(set(found_tickers)):
            raise ValueError(
                "All tickers specified in `tickers` must be in `df`."
                f"Missing tickers: {set(tickers) - set(found_tickers)}"
            )

    else:
        assert bcidxcats, "Failed to resolve tickers"

    if bcidxcats:
        for lx, nm in zip([cids, xcats], ["cid", "xcat"]):
            if not (isinstance(lx, list) and all(isinstance(x, str) for x in lx)):
                raise TypeError(f"`{nm}` must be a list of strings")
            if not set(lx).issubset(set(df[nm])):
                raise ValueError(
                    f"All '{nm}s' in `{nm}` specified must be in `df`. "
                    f"Missing {nm}s: {set(lx) - set(df[nm])}."
                )

        tks: List[str] = [f"{c}_{x}" for c in cids for x in xcats]
        if not len(tks) == 2:
            raise ValueError(
                "The combination of `cids` & `xcats` must yield two tickers",
                f"Found {len(tks)} tickers: {tks}, ",
                f"from `cids` {cids} and `xcats` {xcats}",
            )
        if not set(tks).issubset(set(found_tickers)):
            raise ValueError(
                "All combinations of `cids` & `xcats` (i.e. tickers) specified must "
                "be in `df`."
                f"Missing tickers: {set(tks) - set(found_tickers)}"
            )

    if not isinstance(freq, str):
        raise TypeError("`freq` must be a string")

    if not isinstance(agg, str):
        raise TypeError("`agg` must be a string")

    if add_constant not in [True, False] or not isinstance(add_constant, bool):
        raise TypeError("`add_constant` must be a boolean")

    return True


if __name__ == "__main__":
    cids: List[str] = ["AUD"]
    xcats: List[str] = ["FX", "EQ"]

    df: pd.DataFrame = make_test_df(
        cids=cids,
        xcats=xcats,
    )

    gct: Dict[Any, Any] = granger_causality_test(
        df=df,
        cids=cids,
        xcats=xcats,
    )

    cids: List[str] = ["AUD", "CAD"]
    xcats: str = "FX"
    # tickers =  AUD_FX, CAD_FX
    df: pd.DataFrame = make_test_df(
        cids=cids,
        xcats=xcats,
    )

    gct: Dict[Any, Any] = granger_causality_test(
        df=df,
        tickers=["AUD_FX", "CAD_FX"],
    )

    print(gct)
